# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_route_table_manager
   :synopsis: module that defines several test cases for the
              route_table_manager module.
   In particular, unit tests for RouteTableManager class.
   Setup : Start RouteTableManager thread instance.
   TearDown : Stop RouteTableManager thread instance.
   RouteTableManager is in charge to maintain the list of Workers
   (TrackerWorker, ExaBGPPeerWorker, BGPPeerWorker) subscriptions, to
   process BGB route events and to dispatch BGP routes to the Workers according
   to their subscriptions.
   Mock class is used to stub Workers.
   Tests are organized as follow :
   - testAx use cases to test worker subscriptions to route target (or match)
     1- with no route to synthesize
     2- with routes to synthesize <=> advertise existing routes to the new
        worker subscription according following rules :
        - route should not be synthesized to its source,
        - route should not be synthesized between BGPPeerWorkers
        - route should not be synthesized if already send to worker for
          another RT
     For both use cases check that (worker, match) is correctly recorded by
     RouteTableManager
     Other test cases : re-subscription to check routes are not synthesized
   - testBx use cases to test worker unsubscriptions to match (without and with
     routes to synthesize) :
     same rules should be applied to generate withdraw events.
     Check (worker, match) has been deleted
     Other test cases : unsubscription to match or by worker not registered
   - testCx to test the processing of route event generated by a worker.
     2 types of route event : advertise (new or update) or withdraw
     1- advertise or withdraw a route entry without event propagation to
        workers: check that entry is recorded/deleted
     2- advertise (new or update) or withdraw with event propagation:
        to test dispatching of route events to the workers according to
        their subscriptions.
     Other test cases : withdraw of a not registered route, advertise of the
     same route (same attr and RTs)
   - testDx : to test worker cleanup
   - testEx : to test dumpState

"""

import mock
import testtools

from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import bgp_peer_worker as bpw
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import route_table_manager as rtm
from networking_bagpipe.bagpipe_bgp.engine import worker
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


MATCH1 = rtm.Match(exa.AFI(exa.AFI.ipv4), exa.SAFI(exa.SAFI.mpls_vpn), t.RT1)
MATCH2 = rtm.Match(exa.AFI(exa.AFI.ipv4), exa.SAFI(exa.SAFI.mpls_vpn), t.RT2)
MATCH3 = rtm.Match(exa.AFI(exa.AFI.ipv4), exa.SAFI(exa.SAFI.mpls_vpn), t.RT3)


class TestRouteTableManager(testtools.TestCase, t.BaseTestBagPipeBGP):

    def setUp(self):
        super(TestRouteTableManager, self).setUp()
        self.rtm = rtm.RouteTableManager(mock.Mock(), mock.Mock())
        self.rtm.start()
        self.set_event_target_worker(self.rtm)

    def tearDown(self):
        super(TestRouteTableManager, self).tearDown()
        self.rtm.stop()
        self.rtm.join()

    def _new_worker(self, worker_name, worker_type):
        worker = mock.Mock(spec=worker_type, name=worker_name)
        worker.name = worker_name
        worker.enqueue = mock.Mock()
        worker._rtm_matches = set()
        worker._rtm_route_entries = set()
        return worker

    def _worker_subscriptions(self, worker, rts,
                              afi=exa.AFI(exa.AFI.ipv4),
                              safi=exa.SAFI(exa.SAFI.mpls_vpn)):
        for rt in rts:
            subscribe = engine.Subscription(afi, safi, rt, worker)
            self.rtm._on_event(subscribe)

    def _worker_unsubscriptions(self, worker, rts,
                                afi=exa.AFI(exa.AFI.ipv4),
                                safi=exa.SAFI(exa.SAFI.mpls_vpn)):
        for rt in rts:
            unsubscribe = engine.Unsubscription(afi, safi, rt, worker)
            self.rtm._on_event(unsubscribe)

    def _check_subscriptions(self, worker, matches):
        for match in matches:
            self.assertIn(match, worker._rtm_matches,
                          "Subscription not found")

    def _check_unsubscriptions(self, worker, matches):
        if '_rtm_matches' not in worker.__dict__:
            return
        for match in matches:
            self.assertNotIn(match, worker._rtm_matches,
                             "Subscription found while it should not: %s" %
                             worker._rtm_matches)

    def _check_events_calls(self, events, advertised_routes, withdrawn_nlris):
        # checks that each advertise event in 'events' is in advertised_routes,
        # that each withdraw event in 'events' is in withdrawn_nlris
        # and that all events in withdrawn_nlris and advertised_routes are in
        # 'events'

        for (call_args, _) in events:
            if (call_args[0].type == engine.RouteEvent.ADVERTISE):
                self.assertIn(call_args[0].route_entry, advertised_routes,
                              "Bad advertised route")
                advertised_routes.remove(call_args[0].route_entry)
            else:  # WITHDRAW
                self.assertIn(call_args[0].route_entry.nlri, withdrawn_nlris,
                              "Bad withdrawn route")
                withdrawn_nlris.remove(call_args[0].route_entry.nlri)
        self.assertEqual(0, len(advertised_routes), "some routes not advert'd")
        self.assertEqual(0, len(withdrawn_nlris), "some routes not withdrawn")

    def test_a1_subscriptions_with_no_route_to_synthesize(self):
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # check subscriptions
        self._check_subscriptions(worker1, [MATCH1, MATCH2])

    def test_a1_check_first_last_local_worker_callback(self):
        bgp_worker1 = self._new_worker("worker.Worker-1", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_worker1, [t.RT1])
        self._wait()
        self.assertEqual(
            0,
            self.rtm.first_local_subscriber_callback.call_count,
            "first_local_subscriber_callback should not have been called "
            " (non local worker)")

        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        self.assertEqual(
            1,
            self.rtm.first_local_subscriber_callback.call_count,
            "first_local_subscriber_callback should have been called")

        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT1])
        self.assertEqual(
            1,
            self.rtm.first_local_subscriber_callback.call_count,
            "first_local_subscriber_callback should not have been called a "
            "second time")

        self._worker_unsubscriptions(worker2, [t.RT1])
        self.assertEqual(
            0,
            self.rtm.last_local_subscriber_callback.call_count,
            "last_local_subscriber_callback should not have been called")

        self._worker_unsubscriptions(worker1, [t.RT1])
        self.assertEqual(
            1,
            self.rtm.last_local_subscriber_callback.call_count,
            "last_local_subscriber_callback should have been called")

        self._worker_unsubscriptions(bgp_worker1, [t.RT1])
        self.assertEqual(
            1,
            self.rtm.last_local_subscriber_callback.call_count,
            "last_local_subscriber_callback should not have been called "
            " (non local worker)")

    def test_a2_subscriptions_with_route_to_synthesize(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI2,
                                     [t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # Worker2 subscribes to RT1
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT1])
        # Worker3 subscribes to RT3
        worker3 = self._new_worker("worker.Worker-3", worker.Worker)
        self._worker_subscriptions(worker3, [t.RT3])
        # BGPPeerWorker2 subscribes to RT1
        bgp_peer_worker2 = self._new_worker("BGPWorker2", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker2, [t.RT1])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check route entry synthesized
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be synthesized to its source")
        self.assertEqual(0, worker3.enqueue.call_count,
                         "no route should be synthesized to Worker3")
        self.assertEqual(0, bgp_peer_worker2.enqueue.call_count,
                         "Route should not be synthesized between BGP workers")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 advertise events should be synthesized to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry], [])
        self.assertEqual(1, worker2.enqueue.call_count,
                         "1 advertise event should be synthesized to Worker2")
        self._check_events_calls(
            worker2.enqueue.call_args_list, [evt1.route_entry], [])

    def test_a3_resubscription(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        route_event = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                            t.NLRI1, [t.RT1, t.RT2],
                                            bgp_peer_worker1, t.NH1)
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # Worker1 subscribes again to RT1
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker2 subscribes to RT1
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT1])
        # Worker1 subscribes again to RT2
        self._worker_subscriptions(worker2, [t.RT2])
        # check route entry synthesized
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [route_event.route_entry], [])
        self.assertEqual(1, worker2.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker2")
        self._check_events_calls(worker2.enqueue.call_args_list,
                                 [route_event.route_entry], [])

    def test_a4_two_subscriptions(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])

        # Worker2 subscribes to RT1
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT1])

        # Worker2 advertises a route to RT1
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                              [t.RT1], worker2, t.NH1)

        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker1")

    def test_b1_unsubscription_with_no_route_to_synthesize(self):
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1, t.RT2])
        # Worker1 unsubscribes to RT1
        self._worker_unsubscriptions(worker1, [t.RT1])
        # BGPPeerWorker1 unsubscribes to RT1 and RT2
        self._worker_unsubscriptions(bgp_peer_worker1, [t.RT1, t.RT2])
        # check subscription/unsubscriptions
        self._check_unsubscriptions(worker1, [MATCH1])
        self._check_subscriptions(worker1, [MATCH2])
        self._check_unsubscriptions(bgp_peer_worker1, [MATCH1, MATCH2])

    def test_b2_unsubscription_with_route_to_synthesize(self):
        # BGPPeerWorker1 advertises a route for RT1
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                     t.NLRI1, [t.RT1, t.RT2],
                                     bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                     t.NLRI2, [t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # Worker2 subscribes to RT2
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT2])
        # Worker3 subscribes to RT3
        worker3 = self._new_worker("worker.Worker-3", worker.Worker)
        self._worker_subscriptions(worker3, [t.RT3])
        # BGPPeerWorker2 subscribes to RT1
        bgp_peer_worker2 = self._new_worker("BGPWorker2", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker2, [t.RT1])
        # Workers and BGPPeerWorker unsubscriptions
        self._worker_unsubscriptions(bgp_peer_worker1, [t.RT1])
        self._worker_unsubscriptions(worker1, [t.RT1])
        self._worker_unsubscriptions(worker2, [t.RT2])
        self._worker_unsubscriptions(worker3, [t.RT3])
        self._worker_unsubscriptions(bgp_peer_worker2, [t.RT1])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check route entry synthesized
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be synthesized to its source")
        self.assertEqual(0, worker3.enqueue.call_count,
                         "no route should be synthesized to Worker3")
        self.assertEqual(0, bgp_peer_worker2.enqueue.call_count,
                         "Route should not be synthesized between "
                         "BGPPeerWorkers")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 advertise event should be synthesized to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry], [])
        self.assertEqual(4, worker2.enqueue.call_count,
                         "4 events should be synthesized to Worker2: "
                         "2 advertise and 2 withdraw")
        self._check_events_calls(worker2.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry],
                                 [evt1.route_entry.nlri,
                                  evt2.route_entry.nlri])

    def test_b3_unsubscription_not_registered(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker1 unsubscribes to RT2
        self._worker_unsubscriptions(worker1, [t.RT2])
        # BGPPeerWorker1 unsubscribes to RT1
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._worker_unsubscriptions(bgp_peer_worker1, [t.RT1, t.RT2])
        # check subscription/unsubscriptions
        self._check_subscriptions(worker1, [MATCH1])
        self._check_unsubscriptions(bgp_peer_worker1, [MATCH1, MATCH2])

    def test_c1_route_advertise_by_worker_without_propagation(self):
        # Worker1 advertises a route for RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        route_event = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                            t.NLRI1, [t.RT1, t.RT2],
                                            worker1, t.NH1)
        # check route entry has been inserted
        self.assertIn(route_event.route_entry, worker1._rtm_route_entries,
                      "Route entry not found")

    def test_c2_route_withdraw_by_worker_without_propagation(self):
        # Worker1 advertises then withdraws a route
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                              [t.RT1, t.RT2], worker1, t.NH1)
        route_event = self._new_route_event(engine.RouteEvent.WITHDRAW,
                                            t.NLRI1, [t.RT1], worker1, t.NH1)
        # check route entry has been removed
        self.assertNotIn(route_event.route_entry, worker1._rtm_route_entries,
                         "Route entry found")

    def test_c3_route_advertise_by_bgp_peer_with_propagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker2 subscribes to RT2
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1, t.RT2])
        # BGPPeerWorker2 subscribes to RT1 and RT2
        bgp_peer_worker2 = self._new_worker("BGPWorker2", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker2, [t.RT1, t.RT2])
        # BGPPeerWorker1 advertises a route for RT1
        route_event = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                            t.NLRI1, [t.RT1], bgp_peer_worker1,
                                            t.NH1)
        # check route_event propagation
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route should be propagated to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [route_event.route_entry], [])
        self.assertEqual(0, worker2.enqueue.call_count,
                         "no route should be propagated to Worker2")
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(0, bgp_peer_worker2.enqueue.call_count,
                         "Route should not be propagated between BGP workers")

    def test_c4_route_withdraw_by_peer_worker_with_propagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker2 subscribes to RT2
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT2])
        # Worker3 subscribes to RT3
        worker3 = self._new_worker("worker.Worker-3", worker.Worker)
        self._worker_subscriptions(worker3, [t.RT3])
        # BGPPeerWorker1 subscribes to RT1
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1])
        # BGPPeerWorker2 subscribes to RT2
        bgp_peer_worker2 = self._new_worker("BGPWorker2", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker2, [t.RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        route_eventA = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                             t.NLRI1, [t.RT1, t.RT2],
                                             bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 withdraw previous route (without RT
        route_eventW = self._new_route_event(engine.RouteEvent.WITHDRAW,
                                             t.NLRI1, [],
                                             bgp_peer_worker1, t.NH1)
        # check route_event propagation
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be propagated to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [route_eventA.route_entry],
                                 [route_eventW.route_entry.nlri])
        self.assertEqual(2, worker2.enqueue.call_count,
                         "2 routes should be propagated to Worker2")
        self._check_events_calls(worker2.enqueue.call_args_list,
                                 [route_eventA.route_entry],
                                 [route_eventW.route_entry.nlri])
        self.assertEqual(0, worker3.enqueue.call_count,
                         "No route should be propagated to Worker3")
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(0, bgp_peer_worker2.enqueue.call_count,
                         "Route should not be propagated between BGP workers")

    def test_c5_route_update_by_bgp_peer_with_withdraw_propagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker2 subscribes to RT2
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT2])
        # Worker3 subscribes to RT3
        worker3 = self._new_worker("worker.Worker-3", worker.Worker)
        self._worker_subscriptions(worker3, [t.RT3])
        # BGPPeerWorker1 advertises a route for RT1, RT2 and RT3
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2, t.RT3],
                                     bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises the same nlri with attributes NH and RTs
        # modification
        evt2 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2],
                                     bgp_peer_worker1, t.NH2)
        # check route event propagation
        # TO DO : check route_event.replaced_route
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be advertised to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry], [])
        self.assertEqual(2, worker2.enqueue.call_count,
                         "2 routes should be advertised to Worker2")
        self._check_events_calls(worker2.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry], [])
        self.assertEqual(2, worker3.enqueue.call_count,
                         "2 routes should be advertised/withdrawn to Worker3")
        self._check_events_calls(worker3.enqueue.call_args_list,
                                 [evt1.route_entry], [evt1.route_entry.nlri])

    def test_c6_route_readvertised(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2, t.RT3])
        # BGPPeerWorker1 advertises a route for RT1, RT2 and RT3
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises the same nlri with same attributes and RTs
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                              [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # check route event propagation
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(1, worker1.enqueue.call_count,
                         "only 1 route should be advertised to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry], [])

    def test_c7_route_withdraw_not_registered(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 withdraw a not registered route (without RT
        self._new_route_event(engine.RouteEvent.WITHDRAW, t.NLRI2, [],
                              bgp_peer_worker1, t.NH1)
        # Waiting for RouteTableManager thread finishes to process route_event
        self._wait()
        # check route_event propagation
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route1 should be propagated to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry], [])
        self.assertEqual(0, bgp_peer_worker1.enqueue.call_count,
                         "Route should not be propagated back to its source")

    def test_d1_worker_cleanup(self):
        # Worker1 subscribes to RT1
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1])
        # Worker2 subscribes to RT2
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1, t.RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        evt1 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                                     [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI2,
                                     [t.RT2], bgp_peer_worker1, t.NH1)
        # Cleanup Worker1
        self.rtm.enqueue(engine.WorkerCleanupEvent(bgp_peer_worker1))
        # Waiting for RouteTableManager thread finishes to process the
        # subscriptions
        self._wait()

        self.assertEqual(
            0,
            self.rtm.last_local_subscriber_callback.call_count,
            "last_local_subscriber_callback should not have been called "
            " (non local worker)")

        # check unsubscriptions
        self._check_unsubscriptions(bgp_peer_worker1, [MATCH1, MATCH2])
        # Check route synthesize to Worker1 and Worker2
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be advert/withdraw to Worker1")
        self._check_events_calls(worker1.enqueue.call_args_list,
                                 [evt1.route_entry], [evt1.route_entry.nlri])
        self.assertEqual(4, worker2.enqueue.call_count,
                         "4 routes should be advert/withdraw to Worker2")
        self._check_events_calls(worker2.enqueue.call_args_list,
                                 [evt1.route_entry, evt2.route_entry],
                                 [evt1.route_entry.nlri,
                                  evt2.route_entry.nlri])
        # Check route entries have been removed for BGPPeerWorker1
        self.assertNotIn(evt1.route_entry, bgp_peer_worker1._rtm_route_entries,
                         "Route entry found")
        self.assertNotIn(evt2.route_entry, bgp_peer_worker1._rtm_route_entries,
                         "Route entry found")

        # Cleanup Worker1
        self.rtm.enqueue(engine.WorkerCleanupEvent(worker1))

        self._wait()
        # check that last local subscriber callback for RT1 is called
        self.assertEqual(1, self.rtm.last_local_subscriber_callback.call_count)

    def test_e1_dump_state(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgp_peer_worker1 = self._new_worker("BGPWorker1", bpw.BGPPeerWorker)
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                              [t.RT1, t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI2,
                              [t.RT2], bgp_peer_worker1, t.NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._worker_subscriptions(bgp_peer_worker1, [t.RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._new_worker("worker.Worker-1", worker.Worker)
        self._worker_subscriptions(worker1, [t.RT1, t.RT2])
        # Worker2 subscribes to RT1
        worker2 = self._new_worker("worker.Worker-2", worker.Worker)
        self._worker_subscriptions(worker2, [t.RT1])
        # Worker3 subscribes to RT3
        worker3 = self._new_worker("worker.Worker-3", worker.Worker)
        self._worker_subscriptions(worker3, [t.RT3])

        self.rtm._dump_state()

    def test_7_matches(self):
        m1a = rtm.Match(exa.AFI(exa.AFI.ipv4),
                        exa.SAFI(exa.SAFI.mpls_vpn),
                        exa.RouteTarget(64512, 1))
        m1b = rtm.Match(exa.AFI(exa.AFI.ipv4),
                        exa.SAFI(exa.SAFI.mpls_vpn),
                        exa.RouteTarget(64512, 1))
        m1c = rtm.Match(exa.AFI(exa.AFI.ipv4),
                        exa.SAFI(exa.SAFI.mpls_vpn),
                        exa.RouteTarget(64512, 1, False))
        m2 = rtm.Match(exa.AFI(exa.AFI.ipv4),
                       exa.SAFI(exa.SAFI.mpls_vpn),
                       exa.RouteTarget(64512, 2))
        m3 = rtm.Match(exa.AFI(exa.AFI.ipv4),
                       exa.SAFI(exa.SAFI.mpls_vpn),
                       exa.RouteTarget(64513, 1))

        self.assertEqual(hash(m1a), hash(m1b))
        self.assertEqual(hash(m1a), hash(m1c))
        self.assertNotEqual(hash(m1a), hash(m2))
        self.assertNotEqual(hash(m1a), hash(m3))

        self.assertEqual(m1a, m1b)
        self.assertEqual(m1a, m1c)
        self.assertNotEqual(m1a, m2)
        self.assertNotEqual(m1a, m3)

    def test_f1_test_empty_rt(self):
        # worker advertises a route with no RT

        w1 = self._new_worker("Worker1", worker.Worker)

        subscribe = engine.Subscription(exa.AFI(exa.AFI.ipv4),
                                        exa.SAFI(exa.SAFI.mpls_vpn),
                                        None, w1)
        self.rtm.enqueue(subscribe)

        w2 = self._new_worker("Worker2", worker.Worker)

        route_event = engine.RouteEvent(
            engine.RouteEvent.ADVERTISE,
            engine.RouteEntry(t.NLRI1, None, exa.Attributes()),
            w2)

        self.rtm.enqueue(route_event)

        self._wait()

        self.assertEqual(1, w1.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker1")



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import testtools

from networking_bagpipe.bagpipe_bgp.api import api
from networking_bagpipe.tests.unit.bagpipe_bgp import base


class TestAPI(base.TestCase):

    @testtools.skipIf(six.PY3, 'dataplane driver init fails under py3')
    def test_api_init(self):
        # instantiate the API, will fail if an exception is raised
        api.PecanAPI()



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import networking_bgpvpn

from neutron.tests.fullstack.resources import config as neutron_cfg

BGPVPN_SERVICE = 'bgpvpn'

BGPVPN_PROVIDER = ('BGPVPN:BaGPipe:networking_bgpvpn.neutron.services.'
                   'service_drivers.bagpipe.bagpipe.BaGPipeBGPVPNDriver:'
                   'default')


class NeutronConfigFixture(neutron_cfg.NeutronConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir,
                 connection, rabbitmq_environment):
        super(NeutronConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, connection, rabbitmq_environment)

        if env_desc.bgpvpn:
            self.config['oslo_policy']['policy_dirs'] = (
                os.path.join(networking_bgpvpn.__path__[0],
                             '..', 'etc', 'neutron', 'policy.d')
            )

            # for L2 BGPVPN tests, we need multiple subnet resources using
            # a common IP subnet
            self.config['DEFAULT'].update({
                'allow_overlapping_ips': True
            })


class BGPVPNProviderConfigFixture(neutron_cfg.ConfigFixture):
    def __init__(self, env_desc, host_desc, temp_dir):
        super(BGPVPNProviderConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename='networking_bgpvpn.conf')

        self.config.update({
            'service_providers': {
                'service_provider': BGPVPN_PROVIDER
            }
        })



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import ipvpn


TEST_RD = exa.RouteDistinguisher.fromElements("42.42.42.42", 5)


def _create_test_ipvpn_nlri(label, nexthop):
    return ipvpn.IPVPNRouteFactory(exa.AFI(exa.AFI.ipv4),
                                   "1.1.1.1/32", label, TEST_RD, nexthop)


class TestNLRIs(unittest.TestCase):

    def setUp(self):
        super(TestNLRIs, self).setUp()

    def tearDown(self):
        super(TestNLRIs, self).tearDown()

    # tests on MPLS VPN NLRIs

    def test_0_mpls_vpn_hash_equal(self):
        # Two indistinct VPN NLRI should
        # hash to the same value, and be equal

        nlri1 = _create_test_ipvpn_nlri(42, "45.45.45.45")
        nlri2 = _create_test_ipvpn_nlri(42, "45.45.45.45")

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)

    def test_1_mpls_vpn_hash_equal(self):
        # Two VPN NLRI distinct only by their *label* should
        # hash to the same value, and be equal

        nlri1 = _create_test_ipvpn_nlri(42, "45.45.45.45")
        nlri2 = _create_test_ipvpn_nlri(0, "45.45.45.45")

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)

    def test_2_mpls_vpn_hash_equal(self):
        # Two VPN NLRI distinct only by their *nexthop* should
        # hash to the same value, and be equal

        nlri1 = _create_test_ipvpn_nlri(42, "45.45.45.45")
        nlri2 = _create_test_ipvpn_nlri(42, "77.77.77.77")

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)

    def test_3_mpls_vpn_hash_equal(self):
        # Two VPN NLRI distinct only by their *action* should
        # hash to the same value, and be equal

        nlri1 = _create_test_ipvpn_nlri(42, "45.45.45.45")
        nlri1.action = exa.OUT.ANNOUNCE

        nlri2 = _create_test_ipvpn_nlri(42, "45.45.45.45")
        nlri2.action = exa.OUT.WITHDRAW

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import random
import threading
import time

from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import worker


INIT = "Init-Event"
CONNECT_NOW = "Connect-Now"
SEND_KEEP_ALIVE = "Send-KeepAlive"
KEEP_ALIVE_RECEIVED = "KeepAlive-received"

DEFAULT_HOLDTIME = 180

ERROR_RETRY_TIMER = 20
KA_EXPIRY_RETRY_TIMER = 5
CONNECT_RETRY_TIMER = 5


class FSM(object):

    '''Represents the state of the BGP Finite State Machine'''

    Idle = "Idle"
    Connect = 'Connect'
    OpenSent = 'OpenSent'
    OpenConfirm = 'OpenConfirm'
    Active = 'Active'
    Established = 'Established'

    def __init__(self, worker):
        self.worker = worker
        self._state = FSM.Idle
        self._prev_state = None
        self.all_states = [FSM.Idle, FSM.Connect, FSM.OpenSent,
                           FSM.OpenConfirm, FSM.Active, FSM.Established]
        self.last_transition_time = time.time()

    @property
    def state(self):
        return self._state

    @property
    def previous_state(self):
        return self._prev_state

    @state.setter
    def state(self, state):
        if state == self._state:
            return
        if state in self.all_states:
            self._prev_state = self._state
            self._state = state
            self.last_transition_time = time.time()
            self.worker.log.info(
                "%s BGP FSM transitioned from '%s' to '%s' state" %
                (self.worker, self._prev_state, self._state))
        else:
            raise Exception("no such state (%s)" % repr(state))

    def __repr__(self):
        return self._state


class StoppedException(Exception):
    pass


class InitiateConnectionException(Exception):
    pass


class OpenWaitTimeout(Exception):
    pass


class ToIdle(object):

    def __init__(self, delay):
        # add 50% random delay to avoid reconnect bursts
        self.delay = delay*random.uniform(1, 1.5)

    def __repr__(self):
        return "ToIdle(%s)" % self.delay


@six.add_metaclass(abc.ABCMeta)
class BGPPeerWorker(worker.Worker,
                    threading.Thread,
                    lg.LookingGlassLocalLogger):

    '''Partially abstract class for a Worker implementing the BGP protocol.'''

    def __init__(self, bgp_manager, peer_address):
        # call super
        threading.Thread.__init__(self)
        self.setDaemon(True)
        worker.Worker.__init__(self, bgp_manager, "BGP-%s" % peer_address)

        self.peer_address = peer_address

        # its up to subclasses to call _set_hold_time again to set holdtime
        # based on value advertized by peer
        self._set_hold_time(DEFAULT_HOLDTIME)

        # used to stop receive_thread
        self._stop_loops = threading.Event()
        # used to track that we've been told to stop:
        self.should_stop = False

        self.send_ka_timer = None
        self.ka_reception_timer = None

        lg.LookingGlassLocalLogger.__init__(self,
                                            self.peer_address.replace(".",
                                                                      "-"))

        self.fsm = FSM(self)

        self.log.debug("INIT %s", self.name)
        self.enqueue(CONNECT_NOW)

    def stop(self):
        super(BGPPeerWorker, self).stop()
        self.should_stop = True
        self._stop_and_clean()

    def _set_hold_time(self, holdtime):
        '''sets the session holdtime

        holdtime in seconds
        keepalive expected, or sent, every holdtime/3 second
        '''
        assert holdtime > 30
        self.kat_period = int(holdtime / 3.0)
        self.kat_expiry_time = self.kat_period * 3

    # called by _event_queue_processor_loop
    def _on_event(self, event):

        self.log.debug("event: %s", event)

        if event == CONNECT_NOW:
            self._connect()

        elif isinstance(event, ToIdle):
            self._to_idle(event.delay)

        elif isinstance(event, engine.RouteEvent):
            if self.fsm.state == FSM.Established:
                self._send(self._update_for_route_event(event))
            else:
                raise Exception("cannot process event in '%s' state"
                                % self.fsm.state)

        elif event == SEND_KEEP_ALIVE:
            self._send(self._keep_alive_message_data())

        elif event == KEEP_ALIVE_RECEIVED:
            self.on_keep_alive_received()

        else:
            self.log.warning("event not processed: %s", event)

    def _stopped(self):
        self.fsm.state = FSM.Idle

    def _connect(self):
        self._reset_local_lg_logs()

        # initiate connection
        self.log.debug("connecting now")

        self.fsm.state = FSM.Connect

        try:
            self._initiate_connection()
        except (InitiateConnectionException, OpenWaitTimeout) as e:
            self.log.warning("%s while initiating connection: %s",
                             e.__class__.__name__, e)
            self._to_active()
            return
        except StoppedException:
            self.log.info("Thread stopped during connection init")
            return
        except Exception as e:
            self.log.warning("Exception while initiating connection: %s", e)
            if self.log.isEnabledFor(logging.DEBUG):
                self.log.exception("")
            self._to_active()
            return

        self._stop_loops.clear()

        self.init_send_keep_alive_timer()
        self.init_keep_alive_reception_timer()

        # spawns a receive thread
        self.receive_thread = threading.Thread(target=self._receive_loop,
                                               name=("%s:receive_loop" %
                                                     self.name))
        self.receive_thread.start()

        self._to_established()

    def _to_active(self):
        self.fsm.state = FSM.Active
        self._stop_and_clean()
        self.init_connect_timer(CONNECT_RETRY_TIMER)

    def _to_established(self):
        self.fsm.state = FSM.Established

    def _to_idle(self, delay_before_connect=0):
        self.fsm.state = FSM.Idle
        self._stop_and_clean()

        if delay_before_connect:
            self.init_connect_timer(delay_before_connect)
        else:
            self.enqueue(CONNECT_NOW)

    def _stop_and_clean(self):
        self._stop_loops.set()

        if self.send_ka_timer:
            self.send_ka_timer.cancel()
        if self.ka_reception_timer:
            self.ka_reception_timer.cancel()

        self._cleanup()

    def is_established(self):
        return self.fsm.state == FSM.Established

    def _receive_loop(self):
        self.log.info("Start receive loop")
        self._stop_loops.clear()
        while not self._stop_loops.isSet():
            try:
                loop_result = self._receive_loop_fun()
                if loop_result == 0:
                    self.log.info("receive_loop_fun returned 0, aborting")
                    break
                elif loop_result == 2:
                    self.log.warning("receive_loop_fun returned 2 (error), "
                                     "aborting receive_loop and reinit'ing")
                    # FIXME: use (Worker.)enqueue_high_priority so that
                    # ToIdle is treated before other events
                    self.enqueue(ToIdle(ERROR_RETRY_TIMER))
                    break
                else:
                    # everything went fine
                    pass
            except Exception as e:
                self.log.error("Error: %s (=> aborting receive_loop and "
                               "reinitializing)", e)
                if self.log.isEnabledFor(logging.WARNING):
                    self.log.exception("")
                # FIXME: use (Worker.)enqueue_high_priority so that
                # ToIdle is treated before other events
                self.enqueue(ToIdle(ERROR_RETRY_TIMER))
                break

        self.log.info("End receive loop")

    # Connect retry timer #####
    def init_connect_timer(self, delay):
        self.log.debug("INIT connect timer (%ds)", delay)
        self.connect_timer = threading.Timer(delay, self.enqueue,
                                             [CONNECT_NOW])
        self.connect_timer.name = "%s:connect_timer" % self.name
        self.connect_timer.start()

    # Sending keep-alive's #####

    def init_send_keep_alive_timer(self):
        self.log.debug("INIT Send Keepalive timer (%ds)", self.kat_period)
        self.send_ka_timer = threading.Timer(self.kat_period,
                                             self.send_keep_alive_trigger)
        self.send_ka_timer.name = "%s:send_ka_timer" % self.name
        self.send_ka_timer.start()

    def send_keep_alive_trigger(self):
        self.log.debug("Trigger to send Keepalive")
        self.enqueue(SEND_KEEP_ALIVE)
        self.init_send_keep_alive_timer()

    # Receiving keep-alive's #####

    def init_keep_alive_reception_timer(self):
        self.log.debug(
            "INIT Keepalive reception timer (%ds)", self.kat_expiry_time)
        self.ka_reception_timer = threading.Timer(
            self.kat_expiry_time,
            self.enqueue,
            [ToIdle(KA_EXPIRY_RETRY_TIMER)])
        self.ka_reception_timer.start()

    def on_keep_alive_received(self):
        self.log.debug("Keepalive received")
        self.ka_reception_timer.cancel()
        self.init_keep_alive_reception_timer()

    # Abstract methods

    @abc.abstractmethod
    def _initiate_connection(self):
        '''Abstract method to initiate a connection

        The implementation will initiated the connection to the BGP peer, do
        the initial BGP handshake (send Open, receive Open, send first
        KeepAlive, receive first KeepAlive) and track the intermediate
        FSM states (OpenSent, OpenConfirm).
        '''
        pass

    @abc.abstractmethod
    def _receive_loop_fun(self):
        '''abstract method for internal receive loop

        Return codes:
        - 0: we decided to stop based on stop_loops being set
        - 1: everything went fine, pursue
        - 2: there was an error
        '''
        pass

    @abc.abstractmethod
    def _keep_alive_message_data(self):
        pass

    @abc.abstractmethod
    def _send(self, data):
        pass

    @abc.abstractmethod
    def _update_for_route_event(self, event):
        pass

    # Looking glass hooks ###

    def get_lg_local_info(self, path_prefix):
        return {
            "protocol": {
                "state": self.fsm.state,
                "previous_state": "(%s)" % self.fsm.previous_state,
                "hold_time": self.kat_expiry_time,
                "last_transition_time": time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(self.fsm.last_transition_time))
            }
        }



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from oslo_utils import uuidutils

from networking_bagpipe.agent import bagpipe_bgp_agent
from networking_bagpipe.agent.bgpvpn import constants as bgpvpn_const
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const

from neutron.plugins.ml2.drivers.linuxbridge.agent.common \
    import constants as lnx_agt_constants
from neutron.plugins.ml2.drivers.linuxbridge.agent \
    import linuxbridge_neutron_agent as lnx_agt
from neutron.plugins.ml2.drivers.openvswitch.agent.common \
    import constants as ovs_agt_constants
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_agt
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

from neutron.tests import base
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent \
    import ovs_test_base


PORT10_ID = uuidutils.generate_uuid()
PORT10 = {'id': PORT10_ID,
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '10.0.0.2'}

PORT11 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '10.0.0.3'}

NETWORK1 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '10.0.0.1',
            'segmentation_id': '101'}

PORT20 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '20.0.0.2'}

PORT21 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '20.0.0.3'}

NETWORK2 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '20.0.0.1',
            'segmentation_id': '202'}

ROUTER1 = {'id': uuidutils.generate_uuid()}

port_2_net = {
    PORT10['id']: NETWORK1,
    PORT11['id']: NETWORK1,
    PORT20['id']: NETWORK2,
    PORT21['id']: NETWORK2,
}

LOCAL_VLAN_MAP = {
    NETWORK1['id']: 31,
    NETWORK2['id']: 52
}

BGPVPN_L2_RT10 = {'route_targets': ['BGPVPN_L2:10'],
                  'import_targets': [],
                  'export_targets': []
                  }

BGPVPN_L2_RT20 = {'route_targets': ['BGPVPN_L2:20'],
                  'import_targets': [],
                  'export_targets': []
                  }

BGPVPN_L3_RT100 = {'route_targets': ['BGPVPN_L3:100'],
                   'import_targets': [],
                   'export_targets': []
                   }

BGPVPN_L3_RT200 = {'route_targets': ['BGPVPN_L3:200'],
                   'import_targets': [],
                   'export_targets': []
                   }


class DummyPort(object):
    def __init__(self, network, port, bgpvpn_port=False,
                 evpn=None, ipvpn=None):
        self.id = port['id']
        self.network_id = network['id']
        self.mac_address = port['mac_address']
        self.ip_address = port['ip_address']
        self.gateway_ip = network['gateway_ip']

        if bgpvpn_port:
            if evpn:
                self.l2vpn = copy.deepcopy(evpn)

            if ipvpn:
                self.l3vpn = copy.deepcopy(ipvpn)
        else:
            if evpn:
                self.evpn = copy.deepcopy(evpn)

            if ipvpn:
                self.ipvpn = copy.deepcopy(ipvpn)


class DummyVif(object):
    def __init__(self, ofport, port_name):
        self.ofport = ofport
        self.port_name = port_name


class DummyBGPVPN(object):
    def __init__(self, network, l2vpn=None, l3vpn=None, gateway_mac=None):
        self.id = uuidutils.generate_uuid()
        self.network_id = network['id']

        if l2vpn:
            self.l2vpn = copy.deepcopy(l2vpn)

        if l3vpn:
            self.l3vpn = copy.deepcopy(l3vpn)

        if gateway_mac:
            self.gateway_mac = gateway_mac


class RTList(list):

    def __eq__(self, other):
        return set(self) == set(other)


class BaseTestAgentExtension(object):

    agent_extension_class = None

    DUMMY_VIF10 = None
    DUMMY_VIF11 = None
    DUMMY_VIF20 = None
    DUMMY_VIF21 = None

    def setUp(self):
        self.mocked_bagpipe_agent = mock.Mock(
            spec=bagpipe_bgp_agent.BaGPipeBGPAgent
        )
        self.mocked_bagpipe_agent.do_port_plug = mock.Mock()
        self.mocked_bagpipe_agent.do_port_plug_refresh = mock.Mock()

        patcher = mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent.'
                             'BaGPipeBGPAgent.get_instance',
                             return_value=self.mocked_bagpipe_agent)
        patcher.start()
        self.addCleanup(patcher.stop)

        self.agent_ext = self.agent_extension_class()
        self.connection = mock.Mock()

    def _port_data(self, port, delete=False, admin_state_up=True):
        data = {
            'port_id': port['id']
        }
        if not delete:
            data.update({
                'port_id': port['id'],
                'admin_state_up': admin_state_up,
                'network_id': port_2_net[port['id']]['id'],
                'segmentation_id': port_2_net[port['id']]['segmentation_id'],
                'network_type': 'vxlan',
                'device_owner': 'compute:None',
                'mac_address': port['mac_address'],
                'fixed_ips': [
                    {
                        'ip_address': port['ip_address'],
                    }
                ]
            })
        return data

    def _get_expected_local_port(self, bbgp_vpn_type, network_id, port_id,
                                 detach=False):
        raise NotImplementedError

    def _check_network_info(self, network_id, expected_size,
                            vpn_type=None, vpn_rts=None):
        if expected_size == 0:
            self.assertNotIn(network_id, self.agent_ext.networks_info,
                             "Network %s expected to have no ports left"
                             % network_id)
        else:
            self.assertIn(network_id, self.agent_ext.networks_info)
            network_info = self.agent_ext.networks_info[network_id]
            self.assertEqual(len(network_info.ports), expected_size,
                             "Network ports size not as expected")


class BaseTestLinuxBridgeAgentExtension(base.BaseTestCase,
                                        BaseTestAgentExtension):

    driver_type = lnx_agt_constants.EXTENSION_DRIVER_TYPE

    def setUp(self):
        base.BaseTestCase.setUp(self)
        BaseTestAgentExtension.setUp(self)

        agent_extension_api = mock.Mock()
        self.agent_ext.consume_api(agent_extension_api)
        self.agent_ext.initialize(self.connection,
                                  lnx_agt_constants.EXTENSION_DRIVER_TYPE)

        patcher = mock.patch('neutron.agent.linux.ip_lib.device_exists',
                             return_value=True)
        patcher.start()
        self.addCleanup(patcher.stop)

    def _get_expected_local_port(self, bbgp_vpn_type, network_id, port_id,
                                 detach=False):
        linuxbr = lnx_agt.LinuxBridgeManager.get_bridge_name(network_id)

        if bbgp_vpn_type == bbgp_const.EVPN:
            r = {
                'linuxbr': linuxbr,
                'local_port': {
                    'linuxif': lnx_agt.LinuxBridgeManager.get_tap_device_name(
                        port_id)
                }
            }
            if detach:
                del r['linuxbr']
            return r
        else:  # if bbgp_const.IPVPN:
            return {
                'local_port': {
                    'linuxif': linuxbr
                }
            }


PATCH_INT_TO_MPLS = 5
PATCH_INT_TO_TUN = 7
PATCH_TUN_TO_MPLS = 1
PATCH_TUN_TO_INT = 4
PATCH_MPLS_TO_TUN = 2
PATCH_MPLS_TO_INT = 6

BR_INT_PATCHES = {
    'patch-tun': PATCH_INT_TO_TUN,
    'patch-int-from-mpls': PATCH_INT_TO_MPLS
}

BR_TUN_PATCHES = {
    'patch-int': PATCH_TUN_TO_INT,
    'patch-to-mpls': PATCH_TUN_TO_MPLS
}

BR_MPLS_PATCHES = {
    'patch-from-tun': PATCH_MPLS_TO_TUN,
    'patch-mpls-to-int': PATCH_MPLS_TO_INT
}


def get_port_ofport_tun_br(port_name):
    return BR_TUN_PATCHES[port_name]


def get_port_ofport_int_br(port_name):
    return BR_INT_PATCHES[port_name]


def add_patch_port_mpls(patch, peer):
    return BR_MPLS_PATCHES[patch]


def add_patch_port_tun(patch, peer):
    return BR_TUN_PATCHES[patch]


def add_patch_port_int(patch, peer):
    return BR_INT_PATCHES[patch]


class BaseTestOVSAgentExtension(ovs_test_base.OVSOFCtlTestBase,
                                BaseTestAgentExtension):

    driver_type = ovs_agt_constants.EXTENSION_DRIVER_TYPE

    DUMMY_VIF10 = DummyVif(10, 'VIF10')
    DUMMY_VIF11 = DummyVif(11, 'VIF11')
    DUMMY_VIF20 = DummyVif(20, 'VIF20')
    DUMMY_VIF21 = DummyVif(21, 'VIF21')

    def setUp(self):
        ovs_test_base.OVSOFCtlTestBase.setUp(self)
        BaseTestAgentExtension.setUp(self)

        self.int_br = self.br_int_cls("br-int")
        self.int_br.add_patch_port = mock.Mock(
            side_effect=add_patch_port_int)
        self.int_br.get_port_ofport = mock.Mock(
            side_effect=get_port_ofport_int_br)
        self.int_br.add_flow = mock.Mock()
        self.int_br.delete_flows = mock.Mock()
        self.int_br.use_at_least_protocol = mock.Mock()

        self.tun_br = self.br_tun_cls("br-tun")
        self.tun_br.add_patch_port = mock.Mock(
            side_effect=add_patch_port_tun)
        self.tun_br.get_port_ofport = mock.Mock(
            side_effect=get_port_ofport_tun_br)
        self.tun_br.add_flow = mock.Mock()
        self.tun_br.delete_flows = mock.Mock()

        agent_extension_api = ovs_ext_agt.OVSAgentExtensionAPI(self.int_br,
                                                               self.tun_br)
        self.agent_ext.consume_api(agent_extension_api)

        br_exists_patcher = mock.patch(
            'neutron.agent.common.ovs_lib.OVSBridge.bridge_exists',
            return_value=True)
        br_exists_patcher.start()
        self.addCleanup(br_exists_patcher.stop)

        add_patch_patcher = mock.patch('neutron.agent.common.ovs_lib.OVSBridge'
                                       '.add_patch_port',
                                       side_effect=add_patch_port_mpls)
        add_patch_patcher.start()
        self.addCleanup(add_patch_patcher.stop)

        self.agent_ext.initialize(self.connection,
                                  ovs_agt_constants.EXTENSION_DRIVER_TYPE)

        self.vlan_manager = vlanmanager.LocalVlanManager()
        for net_id, vlan in LOCAL_VLAN_MAP.items():
            try:
                self.vlan_manager.add(net_id, vlan, None, None, None)
            except vlanmanager.MappingAlreadyExists:
                pass

    def _get_expected_local_port(self, bbgp_vpn_type, network_id, port_id,
                                 detach=False):
        vlan = self.vlan_manager.get(network_id).vlan
        if bbgp_vpn_type == bbgp_const.IPVPN:
            r = dict(
                local_port=dict(
                    linuxif='%s:%s' % (bgpvpn_const.LINUXIF_PREFIX, vlan),
                    ovs=dict(plugged=True,
                             port_number=PATCH_MPLS_TO_TUN,
                             vlan=vlan)
                )
            )
            if detach:
                del r['local_port']['ovs']
            return r
        else:
            r = dict(
                local_port=dict(
                    linuxif='%s:%s' % (bgpvpn_const.LINUXIF_PREFIX, vlan),
                    vlan=vlan
                )
            )
            if detach:
                del r['local_port']['vlan']
            return r



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import run_command
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import constants
from networking_bagpipe.bagpipe_bgp.engine import bgp_manager
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import evpn
from networking_bagpipe.bagpipe_bgp.vpn import identifier_allocators
from networking_bagpipe.bagpipe_bgp.vpn import ipvpn

from neutron_lib import exceptions


LOG = logging.getLogger(__name__)

INSTANCE_ID_MAX = 2**32-1


def redirect_instance_extid(instance_type, rt):
    '''generate the ext intance id of a redirection VPN instance'''
    return "redirect-to-%s-%s" % (instance_type, rt.replace(":", "_"))


class MaxInstanceIDReached(exceptions.NeutronException):
    _message = "Could not create VPN instance: max instance id was reached"


class VPNManager(lg.LookingGlassMixin):
    """VPN Manager

    Creates, and keeps track of, VPN instances (VRFs and EVIs) and passes
    plug/unplug calls to the right VPN instance.
    """

    _instance = None

    type2class = {constants.IPVPN: ipvpn.VRF,
                  constants.EVPN: evpn.EVI
                  }

    @log_decorator.log
    def __init__(self):
        LOG.debug("Instantiating VPN Manager...")

        self.bgp_manager = bgp_manager.Manager.get_instance()

        self.dataplane_drivers = dp_drivers.instantiate_dataplane_drivers()

        # VPN instance dict
        self.vpn_instances = {}
        self.next_vpn_instance_id = 1

        LOG.debug("Creating label allocator")
        self.label_allocator = identifier_allocators.LabelAllocator()

        LOG.debug("Creating route distinguisher allocator")
        self.rd_allocator = identifier_allocators.RDAllocator(
            self.bgp_manager.get_local_address())

        # dict containing info how an ipvpn is plugged
        # from an evpn  (keys: ipvpn instances)
        self._evpn_ipvpn_ifs = {}

        self.lock = threading.Lock()

    def _run_command(self, *args, **kwargs):
        run_command.run_command(LOG, *args, run_as_root=True, **kwargs)

    @log_decorator.log_info
    def _attach_evpn_2_ipvpn(self, localport, ipvpn_instance):
        # Assuming localport indicates no real interface but only
        # an EVPN, this method will create a pair of twin interfaces, one
        # to plug in the EVPN, the other to plug in the IPVPN.
        #
        # The localport dict will be modified so that the 'linuxif' indicates
        # the name of the interface to plug in the IPVPN.
        #
        # The EVPN instance will be notified so that it forwards traffic
        # destinated to the gateway on the interface toward the IPVPN.

        assert 'evpn' in localport

        if 'id' not in localport['evpn']:
            raise Exception("Missing parameter 'id' :an external EVPN "
                            "instance id must be specified for an EVPN "
                            "attachment")

        try:
            evpn = self.vpn_instances[localport['evpn']['id']]
        except Exception:
            raise Exception("The specified evpn instance does not exist (%s)"
                            % localport['evpn'])

        if evpn.type != constants.EVPN:
            raise Exception("The specified instance to plug is not an evpn"
                            "instance (is %s instead)" % evpn.type)

        if ipvpn_instance in self._evpn_ipvpn_ifs:
            (evpn_if, ipvpn_if, evpn, managed) = \
                self._evpn_ipvpn_ifs[ipvpn_instance]

            if localport['evpn']['id'] != evpn.external_instance_id:
                raise Exception('Trying to plug into an IPVPN a new E-VPN '
                                'while one is already plugged in')
            else:
                # do nothing
                LOG.warning('Trying to plug an E-VPN into an IPVPN, but it was'
                            ' already done')
                localport['linuxif'] = ipvpn_if
                return

        #  detect if this evpn is already plugged into an IPVPN
        if evpn.has_gateway_port():
            raise Exception("Trying to plug E-VPN into an IPVPN, but this EVPN"
                            " is already plugged into an IPVPN")

        if 'linuxif' in localport and localport['linuxif']:
            raise Exception("Cannot specify an attachment with both a linuxif "
                            "and an evpn")

        if 'ovs_port_name' in localport['evpn']:
            try:
                assert localport['ovs']['plugged']
                assert(localport['ovs']['port_name'] or
                       localport['ovs']['port_number'])
            except Exception:
                raise Exception("Using ovs_port_name in EVPN/IPVPN attachment"
                                " requires specifying the corresponding OVS"
                                " port, which must also be pre-plugged")

            evpn_if = localport['evpn']['ovs_port_name']

            # we assume in this case that the E-VPN interface is already
            # plugged into the E-VPN bridge
            managed = False
        else:
            evpn_if = "evpn%d-ipvpn%d" % (
                evpn.instance_id, ipvpn_instance.instance_id)
            ipvpn_if = "ipvpn%d-evpn%d" % (
                ipvpn_instance.instance_id, evpn.instance_id)

            # FIXME: do it only if not existing already...
            LOG.info("Creating veth pair %s %s ", evpn_if, ipvpn_if)

            # delete the interfaces if they exist already
            self._run_command("ip link delete %s" % evpn_if,
                              acceptable_return_codes=[0, 1])
            self._run_command("ip link delete %s" % ipvpn_if,
                              acceptable_return_codes=[0, 1])

            self._run_command("ip link add %s type veth peer name %s"
                              " mtu 65535" % (evpn_if, ipvpn_if))

            self._run_command("ip link set %s up" % evpn_if)
            self._run_command("ip link set %s up" % ipvpn_if)
            managed = True

        localport['linuxif'] = ipvpn_if

        evpn.set_gateway_port(evpn_if, ipvpn_instance)

        self._evpn_ipvpn_ifs[ipvpn_instance] = (
            evpn_if, ipvpn_if, evpn, managed)

    @log_decorator.log_info
    def _detach_evpn_2_ipvpn(self, ipvpn):
        # Symmetric to _attach_evpn_2_ipvpn

        (evpn_if, ipvpn_if, evpn_instance,
         managed) = self._evpn_ipvpn_ifs[ipvpn]

        if not ipvpn.has_enpoint(ipvpn_if):
            # TODO(tmorin): check that this evpn instance is still up and
            # running ?
            evpn_instance.gateway_port_down(evpn_if)

            # cleanup veth pair
            if managed:
                self._run_command("ip link delete %s" % evpn_if)

            del self._evpn_ipvpn_ifs[ipvpn]

    def _cleanup_evpn2ipvpn(self, ipvpn):
        (_, ipvpn_if, _, managed) = self._evpn_ipvpn_ifs[ipvpn]

        # cleanup veth pair
        if managed:
            self._run_command("ovs-vsctl del-port %s" % ipvpn_if)
            self._run_command("ip link delete %s" % ipvpn_if)

    @utils.synchronized
    @log_decorator.log_info
    def _get_vpn_instance(self, external_instance_id, instance_type,
                          import_rts, export_rts, gateway_ip, mask,
                          readvertise, attract_traffic, fallback=None,
                          **kwargs):
        # Get an vpn_instance with this external_instance_id,
        # if one already exists, check matching instance_type
        # else create one with provided parameters and start it
        #   (unless create_if_none is False --> raise exc.VPNNotFound)
        # returns True if an already started instance was found
        # False if a new instance was created without starting it

        LOG.info("Finding %s for external vpn_instance identifier %s",
                 instance_type, external_instance_id)

        vpn_instance = self.vpn_instances.get(external_instance_id)

        if vpn_instance:
            if vpn_instance.type != instance_type:
                raise Exception("Found an existing vpn_instance with "
                                "external id %s but a different type "
                                "(asked %s vs. already having %s)"
                                % (external_instance_id,
                                   instance_type, vpn_instance.type))
            return vpn_instance, True

        if not kwargs.pop('create_if_none', True):
            raise exc.VPNNotFound(external_instance_id)

        vpn_instance_class = VPNManager.type2class[instance_type]
        dataplane_driver = self.dataplane_drivers[instance_type]

        # unique internal vpn instance id
        instance_id = self.next_vpn_instance_id
        if instance_id > INSTANCE_ID_MAX:
            raise MaxInstanceIDReached()

        self.next_vpn_instance_id += 1

        vpn_instance = vpn_instance_class(self, dataplane_driver,
                                          external_instance_id, instance_id,
                                          import_rts, export_rts,
                                          gateway_ip, mask,
                                          readvertise, attract_traffic,
                                          fallback, **kwargs)

        self.vpn_instances[external_instance_id] = vpn_instance

        return vpn_instance, False

    @utils.synchronized
    @log_decorator.log_info
    def remove_from_vpn_instances(self, external_instance_id):
        del self.vpn_instances[external_instance_id]

    def _check_instance_type(self, params):
        if 'vpn_type' not in params:
            raise exc.APIException("missing instance_type")

        instance_type = params['vpn_type']
        if instance_type not in self.type2class:
            raise exc.APIException("unknown vpn_type: %s" % instance_type)

        if instance_type not in self.dataplane_drivers:
            LOG.error("No dataplane driver for VPN type %s", instance_type)
            raise exc.APIException("No dataplane driver for VPN type %s" %
                                   instance_type)

        return instance_type

    @log_decorator.log_info
    def plug_vif_to_vpn(self, **params):

        instance_type = self._check_instance_type(params)

        vpn_instance_class = VPNManager.type2class[instance_type]
        vpn_instance_class.validate_convert_attach_params(params)

        external_instance_id = params.get('external_instance_id')
        import_rts = params.get('import_rts')
        export_rts = params.get('export_rts')
        mac_address = params.get('mac_address')
        gateway_ip = params.get('gateway_ip')
        localport = params.get('localport')
        linuxbr = params.get('linuxbr')
        advertise_subnet = params.get('advertise_subnet')
        readvertise = params.get('readvertise')
        attract_traffic = params.get('attract_traffic')
        lb_consistent_hash_order = params.get('lb_consistent_hash_order')
        local_pref = params.get('local_pref')
        fallback = params.get('fallback')
        vni = params.get('vni')

        ip_address_prefix = params.get('ip_address_prefix')
        ip_address_plen = params.get('ip_address_plen')

        # Convert route target string to RouteTarget dictionary
        import_rts = utils.convert_route_targets(import_rts)
        export_rts = utils.convert_route_targets(export_rts)

        if readvertise:
            try:
                readvertise = {k: utils.convert_route_targets(readvertise[k])
                               for k in ['from_rt', 'to_rt']}
            except KeyError as e:
                raise Exception("Wrong 'readvertise' parameters: %s" % e)

        if attract_traffic:
            try:
                attract_traffic['redirect_rts'] = (
                    utils.convert_route_targets(
                        attract_traffic['redirect_rts'])
                )
            except KeyError as e:
                raise Exception("Wrong 'attract_traffic' parameters: %s" % e)

        kwargs = {}
        if vni:
            kwargs['vni'] = vni
        if instance_type == constants.EVPN and linuxbr:
            kwargs['linuxbr'] = linuxbr

        vpn_instance, started = self._get_vpn_instance(
            external_instance_id, instance_type, import_rts, export_rts,
            gateway_ip, ip_address_plen, readvertise, attract_traffic,
            fallback, **kwargs)

        vpn_instance.description = params.get('instance_description')

        vpn_instance.update_route_targets(import_rts, export_rts)
        vpn_instance.update_fallback(fallback)

        if instance_type == constants.IPVPN and 'evpn' in localport:
            # special processing for the case where what we plug into
            # the ipvpn is not an existing interface but an interface
            # to create, connected to an existing evpn instance
            self._attach_evpn_2_ipvpn(localport, vpn_instance)

        plug_kwargs = {}
        plug_kwargs['description'] = params.get('description')
        plug_kwargs['direction'] = params.get('direction')

        # Plug VIF to VPN instance
        vpn_instance.vif_plugged(mac_address, ip_address_prefix, localport,
                                 advertise_subnet, lb_consistent_hash_order,
                                 local_pref, **plug_kwargs)

        # delaying the start after the first vif_plugged allows to handle
        # dataplane driver for which the first vif_plugged needs to happen
        # before route advertisements can be processed
        if not started:
            vpn_instance.start()

    @log_decorator.log_info
    def unplug_vif_from_vpn(self, **params):

        instance_type = self._check_instance_type(params)

        vpn_instance_class = VPNManager.type2class[instance_type]
        vpn_instance_class.validate_convert_detach_params(params)

        external_instance_id = params.get('external_instance_id')
        mac_address = params.get('mac_address')
        localport = params.get('localport')
        ip_address_prefix = params.get('ip_address_prefix')

        # Retrieve VPN instance or raise exception if does not exist
        try:
            vpn_instance = self.vpn_instances[external_instance_id]
        except KeyError:
            LOG.error("Try to unplug VIF from non existing VPN instance %s",
                      external_instance_id)
            raise exc.VPNNotFound(external_instance_id)

        # Unplug VIF from VPN instance
        vpn_instance.vif_unplugged(mac_address, ip_address_prefix)

        if vpn_instance.type == constants.IPVPN and 'evpn' in localport:
            self._detach_evpn_2_ipvpn(vpn_instance)

        if vpn_instance.stop_if_empty():
            self.remove_from_vpn_instances(external_instance_id)

    def redirect_instance_for_rt(self, redirected_type, redirect_rt,
                                 stop=False):
        external_instance_id = redirect_instance_extid(redirected_type,
                                                       redirect_rt)
        LOG.info("Need VPN instance %s for traffic redirection to route "
                 "target %s", external_instance_id, redirect_rt)

        # Convert route target string to RouteTarget dictionary
        import_rts = utils.convert_route_targets([redirect_rt])

        # Retrieve a redirect VPN instance or create a new one if none exists
        # yet
        try:
            i, s = self._get_vpn_instance(external_instance_id,
                                          redirected_type,
                                          import_rts, [],
                                          "127.0.0.1",
                                          "24", None, None,
                                          create_if_none=(not stop))
            if not s:
                i.start()
            return i
        except exc.VPNNotFound:
            # (reached only in the 'stop' case)
            LOG.error("Try to stop traffic redirection for an RT for which"
                      " no VPN instance exists (%s)", external_instance_id)
            raise

    @log_decorator.log_info
    def redirect_traffic_to_vpn(self, redirected_id,
                                redirected_type, redirect_rt):
        redirect_instance = self.redirect_instance_for_rt(redirected_type,
                                                          redirect_rt)
        redirect_instance.register_redirected_instance(redirected_id)
        return redirect_instance

    @log_decorator.log_info
    def stop_redirect_to_vpn(self, redirected_id,
                             redirected_type, redirect_rt):
        redirect_instance = self.redirect_instance_for_rt(redirected_type,
                                                          redirect_rt,
                                                          stop=True)
        redirect_instance.unregister_redirected_instance(redirected_id)

        if redirect_instance.stop_if_no_redirected_instance():
            self.remove_from_vpn_instances(
                redirect_instance.external_instance_id)

    @log_decorator.log_info
    def stop(self):
        self.bgp_manager.stop()
        for vpn_instance in six.itervalues(self.vpn_instances):
            vpn_instance.stop()
            # Cleanup veth pair
            if (vpn_instance.type == constants.IPVPN and
                    self._evpn_ipvpn_ifs.get(vpn_instance)):
                self._cleanup_evpn2ipvpn(vpn_instance)
        for vpn_instance in six.itervalues(self.vpn_instances):
            vpn_instance.join()
        self.vpn_instances.clear()

    @classmethod
    @utils.oslo_synchronized('VPNManager')
    def _create_instance(cls):
        if not cls.has_instance():
            cls._instance = cls()

    @classmethod
    def has_instance(cls):
        return cls._instance is not None

    @classmethod
    def clear_instance(cls):
        cls._instance = None

    @classmethod
    def get_instance(cls):
        # double checked locking
        if not cls.has_instance():
            cls._create_instance()
        return cls._instance

    # Looking Glass hooks ####

    def get_lg_map(self):
        class DataplaneLGHook(lg.LookingGlassMixin):

            def __init__(self, vpn_manager):
                self.manager = vpn_manager

            def get_lg_map(self):
                return {
                    "drivers": (lg.COLLECTION, (
                        self.manager.get_lg_dataplanes_list,
                        self.manager.get_lg_dataplane_from_path_item)),
                    "ids": (lg.DELEGATE, self.manager.label_allocator)
                }
        dataplane_hook = DataplaneLGHook(self)
        return {
            "instances": (lg.COLLECTION, (self.get_lg_vpn_list,
                                          self.get_lg_vpn_from_path_item)),
            "dataplane": (lg.DELEGATE, dataplane_hook)
        }

    def get_lg_vpn_list(self):
        return [instance.get_lg_summary()
                for instance in list(self.vpn_instances.values())]

    def get_lg_vpn_from_path_item(self, path_item):
        return self.vpn_instances[path_item]

    def get_vpn_instances_count(self):
        return len(self.vpn_instances)

    def get_lg_dataplanes_list(self):
        return [{"id": i} for i in self.dataplane_drivers.keys()]

    def get_lg_dataplane_from_path_item(self, path_item):
        return self.dataplane_drivers[path_item]



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import itertools
import six

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import constants
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import flowspec
from networking_bagpipe.bagpipe_bgp.engine import ipvpn as ipvpn_routes
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance

DEFAULT_ADDR_PREFIX = '0.0.0.0/0'


@six.add_metaclass(abc.ABCMeta)
class VPNInstanceDataplane(dp_drivers.VPNInstanceDataplane):

    @abc.abstractmethod
    def add_dataplane_for_traffic_classifier(self, classifier,
                                             redirect_to_instance_id):
        pass

    @abc.abstractmethod
    def remove_dataplane_for_traffic_classifier(self, classifier):
        pass


class DummyVPNInstanceDataplane(dp_drivers.DummyVPNInstanceDataplane,
                                VPNInstanceDataplane):

    def add_dataplane_for_traffic_classifier(self, *args, **kwargs):
        raise Exception("not implemented")

    def remove_dataplane_for_traffic_classifier(self, *args, **kwargs):
        raise Exception("not implemented")


class DummyDataplaneDriver(dp_drivers.DummyDataplaneDriver):
    type = constants.IPVPN
    dataplane_instance_class = DummyVPNInstanceDataplane


class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
    # component managing a VRF:
    # - calling a driver to instantiate the dataplane
    # - registering to receive routes for the needed route targets
    # - calling the driver to setup/update/remove routes in the dataplane
    # - cleanup: calling the driver, unregistering for BGP routes

    type = constants.IPVPN
    afi = exa.AFI.ipv4
    safi = exa.SAFI.mpls_vpn

    @log_decorator.log
    def __init__(self, *args, **kwargs):
        vpn_instance.VPNInstance.__init__(self, *args, **kwargs)
        self.readvertised = set()

    def needs_cleanup_assist(self, afi, safi):
        if (afi, safi) == (self.afi, exa.SAFI.flow_vpn):
            return True
        else:
            return self.dataplane.needs_cleanup_assist()

    @classmethod
    def validate_convert_params(cls, params, also_mandatory=()):
        return super(VRF, cls).validate_convert_params(
            params,
            also_mandatory=tuple(set(also_mandatory) | set(['ip_address'])))

    @classmethod
    def validate_convert_attach_params(cls, params):
        super(VRF, cls).validate_convert_attach_params(params)
        if 'gateway_ip' not in params:
            raise exc.APIMissingParameterException('gateway_ip')

    def _nlri_from(self, prefix, label, rd):
        assert rd is not None

        return ipvpn_routes.IPVPNRouteFactory(
            self.afi, prefix, label, rd,
            self.dp_driver.get_local_address())

    def generate_vif_bgp_route(self, mac_address, ip_prefix, plen, label, rd):
        # Generate BGP route and advertise it...
        nlri = self._nlri_from("%s/%s" % (ip_prefix, plen), label, rd)

        return engine.RouteEntry(nlri)

    def _get_local_labels(self):
        for port_data in self.mac_2_localport_data.values():
            yield port_data['label']

    def _get_route_params_for_endpoint(self, endpoint):
        port_data = self.mac_2_localport_data[endpoint[0]]
        label = port_data['label']
        lb_consistent_hash_order = port_data[
            'lb_consistent_hash_order']
        rd = self.endpoint_2_rd[endpoint]

        return (label, rd, lb_consistent_hash_order)

    def _imported(self, route):
        return len(set(route.route_targets).intersection(set(self.import_rts))
                   ) > 0

    def _to_readvertise(self, route):
        rt_records = route.ecoms(exa.RTRecord)
        self.log.debug("RTRecords: %s (readvertise_to_rts:%s)",
                       rt_records,
                       self.readvertise_to_rts)

        readvertise_targets_as_records = [exa.RTRecord.from_rt(rt)
                                          for rt in self.readvertise_to_rts]

        if self.attract_traffic:
            readvertise_targets_as_records += [exa.RTRecord.from_rt(rt)
                                               for rt in self.attract_rts]

        if set(readvertise_targets_as_records).intersection(set(rt_records)):
            self.log.debug("not to re-advertise because one of the readvertise"
                           " or attract-redirect RTs is in RTRecords: %s",
                           set(readvertise_targets_as_records)
                           .intersection(set(rt_records)))
            return False

        return len(set(route.route_targets).intersection(
            set(self.readvertise_from_rts)
            )) > 0

    def _gen_rtrecords_extended_community(self, ecoms):
        # new RTRecord = original RTRecord (if any) + orig RTs converted into
        # RTRecord attributes
        orig_rtrecords = ecoms(exa.RTRecord)
        rts = ecoms(exa.RTExtCom)
        add_rtrecords = [exa.RTRecord.from_rt(rt) for rt in rts]

        return list(set(orig_rtrecords) | set(add_rtrecords))

    def _route_for_readvertisement(self, route, endpoint):
        label, rd, lb_consistent_hash_order = (
            self._get_route_params_for_endpoint(endpoint)
        )
        self.log.debug("Prefix %s (re-)advertisement with label %s and route "
                       "distinguisher %s", route.nlri.cidr.prefix(), label, rd)
        nlri = self._nlri_from(route.nlri.cidr.prefix(), label, rd)

        attributes = exa.Attributes()

        ecoms = self._gen_encap_extended_communities()
        ecoms.communities += (
            self._gen_rtrecords_extended_community(route.ecoms)
        )
        ecoms.communities.append(
            exa.ConsistentHashSortOrder(lb_consistent_hash_order))
        attributes.add(ecoms)

        entry = engine.RouteEntry(nlri, self.readvertise_to_rts, attributes)
        self.log.debug("RouteEntry for (re-)advertisement: %s", entry)
        return entry

    def _default_route_for_advertisement(self, endpoint):
        label, rd, lb_consistent_hash_order = (
            self._get_route_params_for_endpoint(endpoint)
        )
        self.log.debug("Default route (re-)advertisement with label %s and "
                       "route distinguisher %s", label, rd)
        nlri = self._nlri_from(DEFAULT_ADDR_PREFIX, label, rd)

        attributes = exa.Attributes()

        ecoms = self._gen_encap_extended_communities()
        ecoms.communities.append(
            exa.ConsistentHashSortOrder(lb_consistent_hash_order))
        attributes.add(ecoms)

        entry = engine.RouteEntry(nlri, self.readvertise_to_rts, attributes)
        self.log.debug("RouteEntry for default prefix advertisement: %s",
                       entry)
        return entry

    @log_decorator.log
    def _routes_for_attract_static_dest_prefixes(self, endpoint):
        if not self.attract_static_dest_prefixes:
            return

        label, rd, _ = self._get_route_params_for_endpoint(endpoint)

        for prefix in self.attract_static_dest_prefixes:
            nlri = self._nlri_from(prefix, label, rd)

            entry = engine.RouteEntry(nlri, self.readvertise_to_rts)
            self.log.debug("RouteEntry for attract static destination prefix: "
                           "%s", entry)
            yield entry

    @log_decorator.log
    def _route_for_redirect_prefix(self, prefix):
        prefix_classifier = utils.dict_camelcase_to_underscore(
            self.attract_classifier)
        prefix_classifier['destination_prefix'] = prefix

        traffic_classifier = vpn_instance.TrafficClassifier(
            **prefix_classifier)
        self.log.debug("Advertising prefix %s for redirection based on "
                       "traffic classifier %s", prefix, traffic_classifier)
        rules = traffic_classifier.map_traffic_classifier_2_redirect_rules()

        return self.synthesize_redirect_bgp_route(rules)

    def _redirect_route_for_readvertisement(self, route):
        # Create a FlowSpec NLRI with distinct RD and a copy of rules from
        # FlowSpec route to readvertise
        nlri = flowspec.FlowRouteFactory(self.afi, self.instance_rd)
        nlri.rules = route.nlri.rules

        attributes = exa.Attributes()

        ecoms = exa.ExtendedCommunities()
        ecoms.communities += (
            self._gen_rtrecords_extended_community(route.ecoms)
        )
        assert len(self.attract_rts) == 1
        rt = self.attract_rts[0]
        ecoms.communities.append(
            exa.TrafficRedirect(exa.ASN(int(rt.asn)), int(rt.number))
        )
        attributes.add(ecoms)

        entry = engine.RouteEntry(nlri, self.readvertise_to_rts, attributes)
        self.log.debug("RouteEntry for redirect (re-)advertisement: %s", entry)
        return entry

    @log_decorator.log
    def _readvertise(self, route):
        nlri = route.nlri

        self.log.debug("Start re-advertising %s from VRF", nlri)
        if self.attract_traffic:
            # Start advertising default route only when the first route to
            # readvertise appears
            if not self.readvertised:
                for endpoint in self.all_endpoints():
                    self.log.debug("Start advertising default route from VRF "
                                   "%d to redirection VRF", self.instance_id)
                    self._advertise_route(
                        self._default_route_for_advertisement(endpoint)
                    )

            if isinstance(nlri, flowspec.Flow):
                # Readvertise FlowSpec route
                self._advertise_route(
                    self._redirect_route_for_readvertisement(route)
                )
            else:
                # Advertise FlowSpec route for prefix
                self._advertise_route(
                    self._route_for_redirect_prefix(nlri.cidr.prefix())
                )

        else:
            for endpoint in self.all_endpoints():
                self.log.debug("Start re-advertising %s from endpoint %s",
                               nlri.cidr.prefix(), endpoint)
                self._advertise_route(
                    self._route_for_readvertisement(route, endpoint)
                )

        self.readvertised.add(route)

    @log_decorator.log
    def _readvertise_stop(self, route, last):
        nlri = route.nlri

        if last:
            self.log.debug("Stop re-advertising %s from VRF", nlri)
            if self.attract_traffic:
                # Stop advertising default route only if the withdrawn route is
                # the last of the routes to readvertise
                if len(self.readvertised) == 1:
                    for endpoint in self.all_endpoints():
                        self.log.debug("Stop advertising default route from "
                                       "VRF to redirection VRF")
                        self._withdraw_route(
                            self._default_route_for_advertisement(endpoint)
                        )

                if isinstance(nlri, flowspec.Flow):
                    # Withdraw readvertised FlowSpec route
                    self._withdraw_route(
                        self._redirect_route_for_readvertisement(route)
                    )
                else:
                    # Withdraw FlowSpec route for prefix
                    self._withdraw_route(
                        self._route_for_redirect_prefix(nlri.cidr.prefix())
                    )
            else:
                for endpoint in self.all_endpoints():
                    self.log.debug("Stop re-advertising %s from endpoint %s",
                                   nlri.cidr.prefix(), endpoint)
                    route_entry = (
                        self._route_for_readvertisement(route, endpoint)
                    )
                    self._withdraw_route(route_entry)

        self.readvertised.remove(route)

    @log_decorator.log_info
    def vif_plugged(self, mac_address, ip_address_prefix, localport,
                    advertise_subnet=False, lb_consistent_hash_order=0,
                    local_pref=None, **kwargs):
        super(VRF, self).vif_plugged(mac_address, ip_address_prefix,
                                     localport, advertise_subnet,
                                     lb_consistent_hash_order, local_pref,
                                     **kwargs)

        if vpn_instance.forward_to_port(kwargs.get('direction')):
            endpoint = (mac_address, ip_address_prefix)
            for route in itertools.chain(
                    self.readvertised,
                    self._routes_for_attract_static_dest_prefixes(endpoint)):
                self.log.debug("Re-advertising %s with this port as next hop",
                               route.nlri)
                if self.attract_traffic:
                    self._advertise_route(
                        self._default_route_for_advertisement(endpoint)
                    )

                    if self.has_only_one_endpoint():
                        self._advertise_route(self._route_for_redirect_prefix(
                            route.nlri.cidr.prefix()))
                else:
                    self._advertise_route(
                        self._route_for_readvertisement(route, endpoint)
                    )

    @log_decorator.log_info
    def vif_unplugged(self, mac_address, ip_address_prefix):
        endpoint = (mac_address, ip_address_prefix)
        direction = self.endpoint_2_direction[endpoint]
        if vpn_instance.forward_to_port(direction):
            for route in itertools.chain(
                    self.readvertised,
                    self._routes_for_attract_static_dest_prefixes(endpoint)):
                self.log.debug("Stop re-advertising %s", route.nlri)
                if self.attract_traffic:
                    self._withdraw_route(
                        self._default_route_for_advertisement(endpoint)
                    )

                    if self.has_only_one_endpoint():
                        self._withdraw_route(self._route_for_redirect_prefix(
                            route.nlri.cidr.prefix()))
                else:
                    self._withdraw_route(
                        self._route_for_readvertisement(route, endpoint)
                    )

        super(VRF, self).vif_unplugged(mac_address, ip_address_prefix)

    # Callbacks for BGP route updates (TrackerWorker) ########################

    def route_to_tracked_entry(self, route):
        if isinstance(route.nlri, ipvpn_routes.IPVPN):
            return route.nlri.cidr.prefix()
        elif isinstance(route.nlri, flowspec.Flow):
            return (flowspec.Flow, route.nlri._rules())
        else:
            self.log.error("We should not receive routes of type %s",
                           type(route.nlri))
            return None

    @utils.synchronized
    @log_decorator.log
    def new_best_route(self, entry, new_route):

        if self.readvertise:
            # check if this is a route we need to re-advertise
            self.log.debug("route RTs: %s", new_route.route_targets)
            self.log.debug("readv from RTs: %s", self.readvertise_from_rts)
            if self._to_readvertise(new_route):
                self.log.debug("Need to re-advertise %s", entry)
                self._readvertise(new_route)

        if not self._imported(new_route):
            self.log.debug("No need to setup dataplane for:%s", entry)
            return

        if isinstance(new_route.nlri, flowspec.Flow):
            if len(new_route.ecoms(exa.TrafficRedirect)) == 1:
                traffic_redirect = new_route.ecoms(exa.TrafficRedirect)
                redirect_rt = "%s:%s" % (traffic_redirect[0].asn,
                                         traffic_redirect[0].target)

                self.start_redirect_traffic(redirect_rt,
                                            new_route.nlri.rules)
            else:
                self.log.warning("FlowSpec action or multiple traffic "
                                 "redirect actions not supported: %s",
                                 new_route.ecoms())
        else:
            prefix = entry

            encaps = self._check_encaps(new_route)
            if not encaps:
                return

            assert len(new_route.nlri.labels.labels) == 1

            lb_consistent_hash_order = 0
            if new_route.ecoms(exa.ConsistentHashSortOrder):
                lb_consistent_hash_order = new_route.ecoms(
                    exa.ConsistentHashSortOrder)[0].order

            self.dataplane.setup_dataplane_for_remote_endpoint(
                prefix, new_route.nexthop,
                new_route.nlri.labels.labels[0], new_route.nlri, encaps,
                lb_consistent_hash_order)

    @utils.synchronized
    @log_decorator.log
    def best_route_removed(self, entry, old_route, last):

        if self.readvertise:
            # check if this is a route we were re-advertising
            if self._to_readvertise(old_route):
                self.log.debug("Need to stop re-advertising %s", entry)
                self._readvertise_stop(old_route, last)

        if isinstance(old_route.nlri, flowspec.Flow):
            if self._imported(old_route):
                if len(old_route.ecoms(exa.TrafficRedirect)) == 1:
                    if last:
                        traffic_redirect = old_route.ecoms(
                            exa.TrafficRedirect)
                        redirect_rt = "%s:%s" % (traffic_redirect[0].asn,
                                                 traffic_redirect[0].target)

                        self.stop_redirect_traffic(redirect_rt,
                                                   old_route.nlri.rules)
                else:
                    self.log.warning("FlowSpec action or multiple traffic "
                                     "redirect actions not supported: %s",
                                     old_route.ecoms())
        else:
            prefix = entry

            # NOTE(tmorin): On new best routes, we only trigger dataplane
            # update events after checking with self._imported(...) that the
            # route was imported (and not a route that we receive because the
            # VRF should readvertise ir). On best_route_removed, we can't do
            # that because we could end up in a situation where:
            # - initially import_rts contains RT X
            # - we receive a route for RT X and install dataplane state
            # - the import_rts list is later updated and RT X is not anymore
            # part of the imported RTs, and the VRF unsubscribes from RT X
            # - we receive the best_route_removed callbacks corresponding to
            # the unsubscribe, but since the route is for no RT that is in
            # import_rts, we don't update the dataplane
            # The result would be to fail to remove dataplane state for this
            # route, so we're better not optimizing this case and remove
            # dataplane state, including possibly for routes that we did
            # not install in it.

            if self._skip_route_removal(last):
                self.log.debug("Skipping removal of non-last route because "
                               "dataplane does not want it")
                return

            # if we still have a route with same dataplane properties in
            # best routes, then we don't want to clear the dataplane entry
            if self.equivalent_route_in_best_routes(
                    old_route,
                    lambda r: (r.nexthop, r.nlri.labels.labels[0])):
                self.log.debug("Route for same dataplane is still in best "
                               "routes, skipping removal")
                return

            encaps = self._check_encaps(old_route)
            if not encaps:
                return

            assert len(old_route.nlri.labels.labels) == 1

            lb_consistent_hash_order = 0
            if old_route.ecoms(exa.ConsistentHashSortOrder):
                lb_consistent_hash_order = old_route.ecoms(
                    exa.ConsistentHashSortOrder)[0].order

            self.dataplane.remove_dataplane_for_remote_endpoint(
                prefix, old_route.nexthop,
                old_route.nlri.labels.labels[0], old_route.nlri, encaps,
                lb_consistent_hash_order)

    # Looking glass ###

    def get_lg_map(self):
        return {
            "readvertised": (lg.SUBTREE, self.get_lg_readvertised_routes),
        }

    def get_lg_readvertised_routes(self, path_prefix):
        return [route.get_lg_local_info(path_prefix)
                for route in self.readvertised]



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import sqlalchemy as sa
from sqlalchemy.orm import exc

from sqlalchemy import false
from sqlalchemy import true

from oslo_versionedobjects import fields as obj_fields

from neutron.api.rpc.callbacks import resources
from neutron.db import _model_query as model_query
from neutron.db import models_v2
from neutron.objects import base
from neutron.objects import common_types
from neutron.objects.ports import Port

from networking_bagpipe.db import sfc_db as bagpipe_db
from networking_bagpipe.driver import constants

from networking_sfc.db import sfc_db


@base.NeutronObjectRegistry.register
class BaGPipeChainHop(base.NeutronDbObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    db_model = bagpipe_db.BaGPipeChainHop

    fields = {
        'id': common_types.UUIDField(),
        'project_id': obj_fields.StringField(),
        'rts': obj_fields.ListOfStringsField(),
        'ingress_gw': obj_fields.IPAddressField(),
        'egress_gw': obj_fields.IPAddressField(),
        'ingress_ppg': common_types.UUIDField(nullable=True,
                                              default=None),
        'egress_ppg': common_types.UUIDField(nullable=True,
                                             default=None),
        'ingress_network': common_types.UUIDField(nullable=True,
                                                  default=None),
        'egress_network': common_types.UUIDField(nullable=True,
                                                 default=None),
        'ingress_ports': obj_fields.ListOfStringsField(nullable=True),
        'egress_ports': obj_fields.ListOfStringsField(nullable=True),
        'readv_from_rts': obj_fields.ListOfStringsField(nullable=True,
                                                        default=None),
        'readv_to_rt': obj_fields.StringField(nullable=True,
                                              default=None),
        'attract_to_rt': obj_fields.StringField(nullable=True,
                                                default=None),
        'redirect_rts': obj_fields.ListOfStringsField(nullable=True,
                                                      default=None),
        'classifiers': obj_fields.StringField(nullable=True,
                                              default=None),
        'reverse_hop': obj_fields.BooleanField(default=False),

        'portchain_id': common_types.UUIDField(),
    }

    fields_no_update = ['id', 'project_id']

    primary_keys = ['portchain_id']

    foreign_keys = {
        'Network': {'ingress_network': 'id', 'egress_network': 'id'}
    }

    synthetic_fields = ['ingress_ports',
                        'egress_ports']

    rt_fields = {'rts', 'readv_from_rts', 'redirect_rts'}

    def from_db_object(self, db_obj):
        super(BaGPipeChainHop, self).from_db_object(db_obj)

        self._load_ingress_ports(db_obj)
        self._load_egress_ports(db_obj)

    def _load_ports_by_side(self, side):
        ports = []
        if getattr(self, side + '_ppg'):
            # pylint: disable=no-member
            reverse_side = (side if self.reverse_hop
                            else constants.REVERSE_PORT_SIDE[side])

            ports = (
                [getattr(pp, reverse_side)
                 for pp in model_query.get_collection_query(
                    self.obj_context,
                    sfc_db.PortPair,
                    filters={'portpairgroup_id': [getattr(self,
                                                          side + '_ppg')]})
                 ]
            )
        elif getattr(self, side + '_network'):
            port_objs = (
                Port.get_objects(self.obj_context,
                                 network_id=getattr(self, side + '_network'))
            )
            ports = [port_obj.id for port_obj in port_objs]

        setattr(self, side + '_ports', ports)
        self.obj_reset_changes([side + '_ports'])

    def _load_ingress_ports(self, db_obj=None):
        self._load_ports_by_side(constants.INGRESS)

    def _load_egress_ports(self, db_obj=None):
        self._load_ports_by_side(constants.EGRESS)

    @staticmethod
    def _is_port_in_pp(context, port_id):
        try:
            query = context.session.query(sfc_db.PortPair)
            query = query.filter(
                sa.or_(sfc_db.PortPair.ingress == port_id,
                       sfc_db.PortPair.egress == port_id))
            query.one()
            return True
        except exc.NoResultFound:
            return False

    @staticmethod
    def _get_chain_hops_for_port_by_ppg_side(context, port_id, side):
        reverse_side = constants.REVERSE_PORT_SIDE[side]

        query = context.session.query(bagpipe_db.BaGPipeChainHop)
        query = query.join(
            sfc_db.PortPairGroup,
            sfc_db.PortPairGroup.id ==
            getattr(bagpipe_db.BaGPipeChainHop, side + '_ppg'))
        query = query.join(
            sfc_db.PortPair,
            sfc_db.PortPair.portpairgroup_id == sfc_db.PortPairGroup.id)
        query = query.filter(
            sa.or_(
                sa.and_(getattr(sfc_db.PortPair, reverse_side) == port_id,
                        bagpipe_db.BaGPipeChainHop.reverse_hop == false()),
                sa.and_(getattr(sfc_db.PortPair, side) == port_id,
                        bagpipe_db.BaGPipeChainHop.reverse_hop == true())))

        return query.all()

    @staticmethod
    def _get_chain_hops_for_port_by_network_side(context, port_id, side):
        reverse_side = constants.REVERSE_PORT_SIDE[side]

        query = context.session.query(bagpipe_db.BaGPipeChainHop)
        query = query.join(
            models_v2.Network,
            sa.or_(
                sa.and_(
                    models_v2.Network.id ==
                    getattr(bagpipe_db.BaGPipeChainHop,
                            side + '_network'),
                    bagpipe_db.BaGPipeChainHop.reverse_hop == false()),
                sa.and_(
                    models_v2.Network.id ==
                    getattr(bagpipe_db.BaGPipeChainHop,
                            reverse_side + '_network'),
                    bagpipe_db.BaGPipeChainHop.reverse_hop == true()))
            )
        query = query.join(
            models_v2.Port,
            models_v2.Port.network_id == models_v2.Network.id)
        query = query.filter(models_v2.Port.id == port_id)

        return query.all()

    @classmethod
    def get_chain_hops_for_port_by_side(cls, context, port_id, side):
        db_objs = []
        if cls._is_port_in_pp(context, port_id):
            db_objs += (
                cls._get_chain_hops_for_port_by_ppg_side(context,
                                                         port_id,
                                                         side)
            )
        else:
            db_objs += (
                cls._get_chain_hops_for_port_by_network_side(context,
                                                             port_id,
                                                             side)
            )

        return [cls._load_object(context, db_obj) for db_obj in db_objs]

    @classmethod
    def get_objects(cls, context, _pager=None, validate_filters=True,
                    **kwargs):
        if 'port_id' in kwargs:
            port_id = kwargs.pop('port_id')
            chain_hops = []

            for side in [constants.INGRESS, constants.EGRESS]:
                chain_hops += cls.get_chain_hops_for_port_by_side(context,
                                                                  port_id,
                                                                  side)

            return chain_hops

        return super(BaGPipeChainHop, cls).get_objects(
            context, _pager=_pager, validate_filters=validate_filters,
            **kwargs)

    @classmethod
    def modify_fields_from_db(cls, db_obj):
        fields = super(BaGPipeChainHop, cls).modify_fields_from_db(db_obj)

        for field in cls.rt_fields:
            fields[field] = (fields[field].split(',')
                             if fields.get(field) else [])

        for gw_ip in ['ingress_gw', 'egress_gw']:
            if fields.get(gw_ip) is not None:
                fields[gw_ip] = netaddr.IPAddress(fields[gw_ip])

        return fields

    @classmethod
    def modify_fields_to_db(cls, fields):
        result = super(BaGPipeChainHop, cls).modify_fields_to_db(fields)

        for field in cls.rt_fields:
            if result.get(field):
                result[field] = ','.join(result[field])

        for gw_ip in ['ingress_gw', 'egress_gw']:
            if result.get(gw_ip) is not None:
                result[gw_ip] = cls.filter_to_str(result[gw_ip])

        return result


@base.NeutronObjectRegistry.register
class BaGPipePortHops(base.NeutronObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'port_id': common_types.UUIDField(),
        'ingress_hops': obj_fields.ListOfObjectsField(
            'BaGPipeChainHop',
            nullable=True),
        'egress_hops': obj_fields.ListOfObjectsField(
            'BaGPipeChainHop',
            nullable=True),
        'service_function_parameters': obj_fields.DictOfStringsField(
            nullable=True)
    }

    synthetic_fields = {'ingress_hops',
                        'egress_hops'}

    @classmethod
    def get_object(cls, context, **kwargs):
        port_id = kwargs['port_id']
        ingress_hops = (
            BaGPipeChainHop.get_chain_hops_for_port_by_side(context,
                                                            port_id,
                                                            constants.INGRESS)
        )
        egress_hops = (
            BaGPipeChainHop.get_chain_hops_for_port_by_side(context,
                                                            port_id,
                                                            constants.EGRESS)
        )
        return cls(port_id=port_id,
                   ingress_hops=ingress_hops,
                   egress_hops=egress_hops)

    @classmethod
    def get_objects(cls, context, **kwargs):
        raise NotImplementedError()


resources.register_resource_class(BaGPipeChainHop)
resources.register_resource_class(BaGPipePortHops)



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from networking_bagpipe.tests.common import json_fixtures

from neutron.common import utils
from neutron.tests.fullstack.resources import config as neutron_cfg


ROOTWRAP_DAEMON_CMD_DFLT = ("sudo /usr/local/bin/oslo-rootwrap-daemon "
                            "/etc/bagpipe-bgp/rootwrap.conf")


def bagpipe_agent_config_fixture_init_common(self, env_desc, host_desc,
                                             local_ip):
    agent_config = self.config.get('agent')
    if agent_config is None:
        self.config.update({'agent': {}})
        agent_config = self.config.get('agent')

    agent_exts = agent_config.get('extensions', '').split(',')

    if env_desc.bagpipe_ml2:
        agent_exts.append('bagpipe')
    if env_desc.bgpvpn:
        agent_exts.append('bagpipe_bgpvpn')

    agent_config.update({
        'extensions': ','.join(filter(None, agent_exts))
    })

    self.config.update({
        'bagpipe': {
            'bagpipe_bgp_ip': local_ip
        }
    })


class LinuxBridgeConfigFixture(neutron_cfg.LinuxBridgeConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, local_ip,
                 physical_device_name):
        super(LinuxBridgeConfigFixture, self).__init__(env_desc, host_desc,
                                                       temp_dir, local_ip,
                                                       physical_device_name)
        bagpipe_agent_config_fixture_init_common(
            self, env_desc, host_desc, local_ip)


class OVSConfigFixture(neutron_cfg.OVSConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, local_ip, mpls_bridge):
        super(OVSConfigFixture, self).__init__(env_desc, host_desc,
                                               temp_dir, local_ip)
        bagpipe_agent_config_fixture_init_common(
            self, env_desc, host_desc, local_ip)

        self.config.bagpipe.update({
            'mpls_bridge': mpls_bridge,
            'tun_to_mpls_peer_patch_port':
                utils.get_rand_device_name(prefix='to-mpls'),
            'mpls_to_tun_peer_patch_port':
                utils.get_rand_device_name(prefix='to-tun'),
            'mpls_to_int_peer_patch_port':
                utils.get_rand_device_name(prefix='mpls-to-int'),
            'int_to_mpls_peer_patch_port':
                utils.get_rand_device_name(prefix='int-to-mpls'),
        })


class JsonFixture(neutron_cfg.ConfigFixture):
    """A fixture that holds a JSON configuration."""
    def _setUp(self):
        cfg_fixture = json_fixtures.JsonFileFixture(
            self.base_filename, self.config, self.temp_dir)
        self.useFixture(cfg_fixture)
        self.filename = cfg_fixture.filename


class BagpipeBGPConfigFixture(neutron_cfg.ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, local_ip, bgp_peer,
                 bgp_port, mpls_bridge, mpls_interface):
        super(BagpipeBGPConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename='bgp.conf')

        self.config.update({
            'DEFAULT': {
                'debug': True,
            },
            'COMMON': {
                'root_helper_daemon': os.environ.get('OS_ROOTWRAP_DAEMON_CMD',
                                                     ROOTWRAP_DAEMON_CMD_DFLT)
            },
            'BGP': {
                'local_address': local_ip,
                'peers': bgp_peer,
                'my_as': '64512',
                'bgp_port': bgp_port,
            },
            'API': {
                'host': local_ip,
                'port': '8082'
            },
            'DATAPLANE_DRIVER_IPVPN': {
                'dataplane_driver': self.env_desc.ipvpn_driver,
                'ovs_bridge': mpls_bridge,
                'proxy_arp': 'False'
            },
            'DATAPLANE_DRIVER_EVPN': {
                'dataplane_driver': self.env_desc.evpn_driver
            }
        })

        if self.env_desc.ipvpn_driver != 'dummy':
            if self.env_desc.ipvpn_encap == 'vxlan':
                self.config['DATAPLANE_DRIVER_IPVPN'].update({
                    'vxlan_encap': 'True',
                    'mpls_interface': ''
                })
            if 'mpls-gre' in self.env_desc.ipvpn_encap:
                self.config['DATAPLANE_DRIVER_IPVPN'].update({
                    'mpls_interface': '*gre*',
                    'gre_tunnel': self._generate_gre_tunnel()
                })
            if self.env_desc.ipvpn_encap == 'mpls-gre-l3':
                self.config['DATAPLANE_DRIVER_IPVPN'].update({
                    'gre_tunnel_options': "packet_type=legacy_l3"
                })
            if self.env_desc.ipvpn_encap == 'bare-mpls':
                self.config['DATAPLANE_DRIVER_IPVPN'].update({
                    'mpls_interface': mpls_interface
                })

    def _generate_gre_tunnel(self):
        return utils.get_rand_device_name(prefix='mpls-gre')


class GoBGPConfigFixture(JsonFixture):

    def __init__(self, env_desc, host_desc, temp_dir,
                 bgp_peer, bgp_port, host_ips):
        super(GoBGPConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename='gobgp.conf')

        self.config.update({
            'global': {
                'config': {
                    'as': '64512',
                    'router-id': bgp_peer,
                    'port': bgp_port
                }
            }
        })

        neighbors = list()
        for host_ip in host_ips:
            neighbor = {
                'config': {
                    'neighbor-address': host_ip,
                    'peer-as': '64512'
                },
                'transport': {
                    'config': {
                        'passive-mode': 'true'
                    }
                },
                'route-reflector': {
                    'config': {
                        'route-reflector-client': 'true',
                        'route-reflector-cluster-id': '1.2.3.4'
                    }
                }
            }

            afi_safis = list()
            for afi_safi in ('rtc', 'l2vpn-evpn', 'l3vpn-ipv4-unicast',
                             'l3vpn-ipv4-flowspec'):
                afi_safis.append({
                    'config': {
                        'afi-safi-name': afi_safi
                    }
                })

            neighbor.update({'afi-safis': afi_safis})
            neighbors.append(neighbor)

        self.config.update({'neighbors': neighbors})



#!/usr/bin/env python
#
# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from sys import stdout

import optparse
from oslo_serialization import jsonutils
import six
import urllib2


BAGPIPE_PORT = 8082
LOOKING_GLASS_BASE = "looking-glass"

INDENT_INCREMENT = 2


def pretty_print_recurse(data, indent, recursive_requests, url,
                         already_anew_line=False):
    """recursively pretty print data

    key has already been output, this function will print data and finish at
    a start of line

    returns True if the key output was spread across multiple lines.
    """
    if isinstance(data, dict):
        more = False

        if ("id" in data and "href" in data):
            stdout.write(data["id"])
            del data["id"]

        if ("href" in data):
            more = True

        if more:
            target_url = data["href"]
            if recursive_requests:
                if target_url.startswith(url):
                    response = urllib2.urlopen(target_url)
                    if response.getcode() == 200:
                        pretty_print_recurse(jsonutils.load(response),
                                             indent + INDENT_INCREMENT,
                                             recursive_requests, target_url,
                                             already_anew_line=False)
                        return True
                    else:
                        stdout.write(" (ERROR %d)", response.getcode())
                        return False

            del data["href"]
            stdout.write(" (...)")
            already_anew_line = False

        if len(data) > 0:
            if not already_anew_line:
                stdout.write("\n")
            first_val = True
            for (key, value) in six.iteritems(data):
                if not first_val or not already_anew_line:
                    stdout.write("%s" % (" " * indent))
                if first_val:
                    first_val = False
                stdout.write("%s: " % key)
                pretty_print_recurse(value, indent + INDENT_INCREMENT,
                                     recursive_requests, url)
        else:
            if more:
                stdout.write("\n")
            else:
                stdout.write("-\n")

    elif isinstance(data, list):

        if len(data) > 0:
            if not already_anew_line:
                stdout.write("\n")
                already_anew_line = True

            for value in data:
                stdout.write("%s* " % (" " * indent))
                if isinstance(value, dict) or isinstance(value, list):
                    pretty_print_recurse(value, indent + INDENT_INCREMENT,
                                         recursive_requests, url,
                                         already_anew_line)
                else:
                    stdout.write("%s\n" % value)
                already_anew_line = True
        else:
            stdout.write("-\n")

    else:
        if isinstance(data, str) and "\n" in data:
            data = data.strip("\n").replace("\n", "\n%s" % (" " * indent))
            stdout.write("\n%s" % (" " * indent))

        stdout.write("%s\n" % data)
        return False


def main():
    usage = """ %prog [--server <ip>] path to object in looking-glass

e.g.: %prog vpns instances"""
    parser = optparse.OptionParser(usage)

    parser.add_option(
        "--server", dest="server", default="127.0.0.1",
        help="IP address of BaGPipe BGP (optional, default: %default)")

    parser.add_option(
        "--port", dest="port", type="int", default=BAGPIPE_PORT,
        help="Port of BaGPipe BGP (optional, default: %default)")

    parser.add_option(
        "--prefix", dest="prefix", default=LOOKING_GLASS_BASE,
        help="Looking-glass URL Prefix (optional, default: %default)")

    parser.add_option(
        "-r", "--recurse", dest="recurse", action="store_true", default=False,
        help="Recurse down into the whole looking-glass (disabled by default)")

    (options, args) = parser.parse_args()

    quoted_args = [urllib2.quote(arg) for arg in args]
    target_url = "http://%s:%d/%s/%s" % (options.server, options.port,
                                         options.prefix, "/".join(quoted_args))
    try:
        os.environ['NO_PROXY'] = options.server
        response = urllib2.urlopen(target_url)

        if response.getcode() == 200:
            data = jsonutils.load(response)

            if (isinstance(data, dict) and "href" in data):
                target_url_bis = data["href"]
                response_bis = urllib2.urlopen(target_url_bis)
                if response.getcode() == 200:
                    target_url = target_url_bis
                    data = jsonutils.load(response_bis)

            pretty_print_recurse(data, 0, options.recurse, target_url,
                                 already_anew_line=True)

    except urllib2.HTTPError as e:
        if e.code == 404:
            print("No such looking glass path: %s\n(%s)" %
                  (" ".join(quoted_args), target_url))
        else:
            print("Error code %d: %s" % (e.getcode(), e.read()))
        return
    except urllib2.URLError as e:
        print("No server at http://%s:%d : %s" % (options.server,
                                                  options.port, e))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from networking_bagpipe.bagpipe_bgp.vpn import identifier_allocators


class TestIDAllocator(testtools.TestCase):

    min_id = 100
    alloc_id_count = 5
    remove_index = 2

    def setUp(self):
        super(TestIDAllocator, self).setUp()

        self.test_allocator = identifier_allocators.IDAllocator()
        self.test_allocator.current_id = self.min_id
        self.test_allocator.MAX = (self.min_id + self.alloc_id_count) - 1

        self.allocated_ids = set()
        for i in range(self.alloc_id_count):
            self.allocated_ids.add(
                self.test_allocator.get_new_id("Test identifier %d" % i)
            )

    def test_allocated_id_uniqueness(self):
        self.assertEqual(self.allocated_ids,
                         set(self.test_allocator.allocated_ids.keys()))
        self.assertEqual(self.test_allocator.MAX + 1,
                         self.test_allocator.current_id)

    def test_reuse_released_id(self):
        # Check identifier reused after having been released
        remove_id = self.min_id + self.remove_index
        self.test_allocator.release(remove_id)

        reused_id = self.test_allocator.get_new_id("Test reused identifier")
        self.assertEqual(remove_id, reused_id)
        self.assertEqual(self.test_allocator.MAX + 1,
                         self.test_allocator.current_id)

    def test_allocated_id_max(self):
        # Check no identifiers left exception
        self.assertRaises(identifier_allocators.MaxIDReached,
                          self.test_allocator.get_new_id,
                          "Test max reached identifier")



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from oslo_log import log as logging
from six.moves import queue

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp import engine


LOG = logging.getLogger(__name__)

STOP_EVENT = "STOP_EVENT"


class Worker(engine.EventSource, lg.LookingGlassMixin):

    """Base class for objects that interact with the route table manager

    These objects will:
    * use _subscribe(...) and _unsubscribe(...) to subscribe to routing events
    * will specialize _on_event(event) to react to received events

    They also inherit from EventSource to publish events
    """

    def __init__(self, bgp_manager, worker_name):
        self.bgp_manager = bgp_manager
        self.rtm = bgp_manager.rtm
        self._queue = queue.Queue()
        self._please_stop = threading.Event()

        self.name = worker_name
        assert self.name is not None

        engine.EventSource.__init__(self, self.rtm)

        # private data for RouteTableManager
        self._rtm_matches = set()

        LOG.debug("Instantiated %s worker", self.name)

    def stop(self):
        """Stop this worker.

        Set the _please_stop internal event to stop the event processor loop
        and indicate to the route table manager that this worker is stopped.
        Then call _stopped() to let a subclass implement any further work.
        """
        LOG.info("Stop worker %s", self)
        self.stop_event_loop()
        self._cleanup()
        self._stopped()

    def stop_event_loop(self):
        self._please_stop.set()
        self.enqueue(STOP_EVENT)

    def _cleanup(self):
        self.rtm.enqueue(engine.WorkerCleanupEvent(self))

    def _stopped(self):
        """Hook for subclasses to react when Worker is stopped

        (NoOp in base Worker class)
        """

    def _event_queue_processor_loop(self):
        """Main loop where the worker consumes events."""

        while not self._please_stop.isSet():
            event = self._dequeue()

            if event == STOP_EVENT:
                LOG.debug("Stop event, breaking queue processor loop")
                self._please_stop.set()
                break

            try:
                self._on_event(event)
            except Exception:
                LOG.exception("Exception raised on subclass._on_event: %s",
                              event)

    def run(self):
        self._event_queue_processor_loop()

    def _on_event(self, event):
        """Method implemented by subclasses to react to routing events."""

        LOG.debug("Worker %s _on_event: %s", self.name, event)
        raise NotImplementedError

    def _dequeue(self):
        return self._queue.get()

    def enqueue(self, event):
        # TODO(tmmorin): replace Queue by a PriorityQueue and use a higher
        # priority for ReInit event
        self._queue.put(event)

    def _subscribe(self, afi, safi, rt=None):
        subobj = engine.Subscription(afi, safi, rt, self)
        LOG.info("Subscribe: %s ", subobj)
        self.rtm.enqueue(subobj)

    def _unsubscribe(self, afi, safi, rt=None):
        subobj = engine.Unsubscription(afi, safi, rt, self)
        LOG.info("Unsubscribe: %s ", subobj)
        self.rtm.enqueue(subobj)

    def get_subscriptions(self):
        return sorted(self._rtm_matches)

    def __repr__(self):
        return "%s" % (self.name)

    # Looking glass ###

    def get_lg_local_info(self, path_prefix):
        return {
            "name": self.name,
            "internals": {
                "event queue length": self._queue.qsize(),
                "subscriptions":
                    [repr(sub) for sub in self.get_subscriptions()],
            }
        }



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections

from oslo_log import log as logging

from networking_bagpipe._i18n import _
from networking_bagpipe.bagpipe_bgp.common import log_decorator

from neutron.agent.common import ovs_lib
from neutron.plugins.ml2.drivers.openvswitch.agent.common import \
    constants as ovs_const

from neutron_lib import exceptions

LOG = logging.getLogger(__name__)


# largely copied from networking_sfc.services.sfc.common.ovs_ext_lib
class OVSBridgeWithGroups(object):

    def __init__(self, ovs_bridge):
        self.bridge = ovs_bridge

        # OpenFlow 1.1 is needed to manipulate groups
        self.bridge.use_at_least_protocol(ovs_const.OPENFLOW11)

    # proxy most methods to self.bridge
    def __getattr__(self, name):
        return getattr(self.bridge, name)

    def do_action_groups(self, action, kwargs_list):
        group_strs = [_build_group_expr_str(kw, action) for kw in kwargs_list]
        if action == 'add' or action == 'del':
            cmd = '%s-groups' % action
        elif action == 'mod':
            cmd = '%s-group' % action
        else:
            msg = _("Action is illegal")
            raise exceptions.InvalidInput(error_message=msg)
        self.run_ofctl(cmd, ['--may-create', '-'], '\n'.join(group_strs))

    @log_decorator.log_info
    def add_group(self, **kwargs):
        self.do_action_groups('add', [kwargs])

    @log_decorator.log_info
    def mod_group(self, **kwargs):
        self.do_action_groups('mod', [kwargs])

    @log_decorator.log_info
    def delete_group(self, **kwargs):
        self.do_action_groups('del', [kwargs])

    def dump_group_for_id(self, group_id):
        retval = None
        group_str = "%d" % group_id
        group = self.run_ofctl("dump-groups", [group_str])
        if group:
            retval = '\n'.join(item for item in group.splitlines()
                               if ovs_lib.is_a_flow_line(item))
        return retval

    def get_bridge_ports(self):
        port_name_list = self.bridge.get_port_name_list()
        of_portno_list = list()
        for port_name in port_name_list:
            of_portno_list.append(self.bridge.get_port_ofport(port_name))
        return of_portno_list


def _build_group_expr_str(group_dict, cmd):
    group_expr_arr = []
    buckets = None
    group_id = None

    if cmd != 'del':
        if "group_id" not in group_dict:
            msg = _("Must specify one group Id on group addition"
                    " or modification")
            raise exceptions.InvalidInput(error_message=msg)
        group_id = "group_id=%s" % group_dict.pop('group_id')

        if "buckets" not in group_dict:
            msg = _("Must specify one or more buckets on group addition"
                    " or modification")
            raise exceptions.InvalidInput(error_message=msg)
        buckets = "%s" % group_dict.pop('buckets')

    if group_id:
        group_expr_arr.append(group_id)

    for key, value in group_dict.items():
        group_expr_arr.append("%s=%s" % (key, value))

    if buckets:
        group_expr_arr.append(buckets)

    return ','.join(group_expr_arr)


class ObjectLifecycleManager(object):

    def __init__(self):
        self.objects = dict()
        self.object_used_for = collections.defaultdict(set)

    @abc.abstractmethod
    def create_object(self, object_key):
        pass

    @abc.abstractmethod
    def delete_object(self, object):
        pass

    @log_decorator.log_info
    def get_object(self, object_key, user_key=None):
        obj = self.objects.get(object_key)
        if obj:
            LOG.debug("existing object for %s: %s", object_key, obj)
        else:
            if not user_key:
                return None

            obj = self.create_object(object_key)
            self.objects[object_key] = obj
            LOG.debug("object for %s: %s", object_key, obj)

        if user_key:
            self.object_used_for[object_key].add(user_key)

        return obj

    @log_decorator.log_info
    def free_object(self, object_key, user_key):
        if object_key not in self.object_used_for:
            LOG.debug("no object to free for %s", object_key)
            return

        self.object_used_for[object_key].discard(user_key)
        if not self.object_used_for[object_key]:
            obj = self.objects[object_key]

            LOG.debug("%s was last user for %s, clearing", user_key,
                      object_key)
            self.delete_object(obj)

            del self.objects[object_key]
            del self.object_used_for[object_key]
        else:
            LOG.debug("remaining users for object %s: %s", object_key,
                      self.object_used_for[object_key])

    def infos(self):
        return self.objects


class ObjectLifecycleManagerProxy(object):

    def __init__(self, manager, parent_user):
        self.manager = manager
        self.parent_user = parent_user

    def get_object(self, object_key, user_key=None):
        if user_key:
            return self.manager.get_object(object_key,
                                           (self.parent_user, user_key))
        else:
            return self.manager.get_object(object_key)

    def free_object(self, object_key, user_key):
        if user_key:
            return self.manager.free_object(object_key,
                                            (self.parent_user, user_key))

    def infos(self):
        return self.manager.infos()
